import heapq

class PriorityQueue(object):
    def __init__(self, content=None):
        if content is None:
            content = []
        self.heap = content
        if self.heap:
            self.update()

    def __iter__(self):
        return self

    def __next__(self):
        return self.next()

    def next(self):
        if not self.heap:
            raise StopIteration
        if len(self.heap) == 1:
            return self.heap.pop()
        last_item = self.heap.pop()
        result = self.heap[0]
        self.heap[0] = last_item
        heapq._siftup_max(self.heap, 0)
        return result

    def add(self, element):
        self.heap.append(element)
        heapq._siftdown_max(self.heap, 0, len(self.heap)-1)

    def update(self):
        heapq._heapify_max(self.heap)

    def __str__(self):
        return str(self.heap)





class BasicBlock(object):

    def __init__(self):
        self.start_addr = 0
        self.end_addr = 0
        self.instructions = []
        self.successors = []

    def __str__(self):
        return "0x%x - 0x%x (%d) -> [%s]" % (self.start_addr, self.end_addr, len(self.instructions), ", ".join(["0x%x" % ref for ref in self.successors]))



#!/usr/bin/python
import re
import lief
lief.logging.disable()
import struct
import logging
from collections import OrderedDict

from .AbstractLabelProvider import AbstractLabelProvider

LOGGER = logging.getLogger(__name__)


class GoSymbolProvider(AbstractLabelProvider):
    """ Minimal resolver for Go symbols """

    def __init__(self, config):
        self._config = config
        # addr:func_name
        self._func_symbols = {}

    def update(self, binary_info):
        binary = binary_info.binary
        is_elf = False
        pclntab_offset = None
        try:
            elf = lief.ELF.parse(bytearray(binary))
            pclntab_offset = elf.get_section(".gopclntab").offset
            is_elf = True
        except:
            pass
        if not is_elf:
            try:
                pe = lief.PE.parse(bytearray(binary))
                rdata_offset = pe.get_section(".rdata").offset
                pclntab_offset = rdata_offset + pe.get_symbol("runtime.pclntab").value
            except:
                pass
        if pclntab_offset is None:
            # scan for offset of structure
            pclntab_regex = re.compile(b".\xFF\xFF\xFF\x00\x00\x01(\x04|\x08)")
            hits = [match.start() for match in re.finditer(pclntab_regex, binary)]
            if len(hits) > 1:
                logging.error("GoLabelProvider found too many candidates for pclntab")
            elif len(hits) == 1:
                pclntab_offset = hits[0]
        # if we found a valid offset, do the pclntab parsing
        if pclntab_offset:
            result = self._parse_pclntab(pclntab_offset, binary)
            if result:
                self._func_symbols = result

    def isSymbolProvider(self):
        return True

    def getSymbol(self, address):
        return self._func_symbols.get(address, "")

    def getFunctionSymbols(self):
        return self._func_symbols

    def _readUtf8(self, buffer):
        string_read = ""
        offset = 0
        while buffer[offset] != 0:
            string_read += f"{buffer[offset]:02x}"
            offset += 1
        # need to defang special char(s)
        decoded_string = bytearray.fromhex(string_read).decode().replace('\u00b7', ':')
        return decoded_string

    def _parse_pclntab(self, pclntab_offset, binary):
        pclntab_buffer = binary[pclntab_offset:]

        marker = struct.unpack("I", pclntab_buffer[0:4])[0]
        if marker == 0xfffffffb:
            version = '1.12'
        elif marker == 0xfffffffa:
            version = '1.16'
        elif marker == 0xfffffff0:
            version = '1.18'
        else:
            raise ValueError(f"Could not recognize Golang version marker: 0x{marker}")
        
        bitness_indicator = struct.unpack("B", pclntab_buffer[7:8])[0]
        bitness = None
        if bitness_indicator == 8:
            bitness = 64
        elif bitness_indicator == 4:
            bitness = 32
        else:
            raise ValueError(f"Could not recognize Golang bitness marker: 0x{bitness_indicator}")

        field_size = 8 if bitness == 64 else 4
        field_indicator = "Q" if bitness == 64 else "I"
        if version == '1.12':
            number_of_functions = struct.unpack("I", pclntab_buffer[8:12])[0]
            function_name_offset = pclntab_offset
            weird_table_offset = pclntab_offset + 16 if bitness == 64 else pclntab_offset + 12
            start_text = 0
        elif version == '1.16':
            parsed_pclntab_fields = struct.unpack(7*field_indicator, pclntab_buffer[8:8+7*field_size])
            number_of_functions = parsed_pclntab_fields[0]
            function_name_offset = pclntab_offset + parsed_pclntab_fields[2]
            file_name_offset = pclntab_offset + parsed_pclntab_fields[3]
            weird_table_offset = pclntab_offset + parsed_pclntab_fields[6]
            start_text = 0
        elif version == '1.18':
            parsed_pclntab_fields = struct.unpack(8*field_indicator, pclntab_buffer[8:8+8*field_size])
            number_of_functions = parsed_pclntab_fields[0]
            start_text = parsed_pclntab_fields[2]
            function_name_offset = pclntab_offset + parsed_pclntab_fields[3]
            file_name_offset = pclntab_offset + parsed_pclntab_fields[5]
            weird_table_offset = pclntab_offset + parsed_pclntab_fields[7]

        # first parse function offsets
        offsets = OrderedDict()
        func_info_offsets = {}
        read_offset = 0
        table_buffer = binary[weird_table_offset:]
        for index in range(number_of_functions):
            # need to parse a second table in this case
            if version == '1.12':
                offsets[index] = struct.unpack(field_indicator, table_buffer[read_offset:read_offset+field_size])[0]
                read_offset += field_size
                func_info_offsets[index] = struct.unpack(field_indicator, table_buffer[read_offset:read_offset+field_size])[0]
                read_offset += field_size
            # advance element pointer
            if version == '1.16':
                offsets[index] = struct.unpack(field_indicator, table_buffer[read_offset:read_offset+field_size])[0]
                read_offset += 2 * field_size
            # here we have a more compact structure for both x86/x64, no need to skip
            if version == '1.18':
                offsets[index] = struct.unpack("I", table_buffer[read_offset:read_offset+4])[0]
                read_offset += 8

        functions = {}
        offsets2 = offsets.copy()
        function_name_buffer = binary[function_name_offset:]
        if version == '1.12':
            for index, info_offset in func_info_offsets.items():
                function_offset = offsets[index]
                name_offset = struct.unpack(field_indicator, pclntab_buffer[info_offset+field_size:info_offset+2*field_size])[0]
                function_name = self._readUtf8(function_name_buffer[name_offset:])
                functions[function_offset + start_text] = function_name
        else:
            delete = False
            for offset, function_offset in offsets.items():
                if delete:
                    offsets2.pop(offset)
                bytes_read = struct.unpack("I", table_buffer[read_offset:read_offset+4])[0]
                read_offset += 4
                try:
                    while bytes_read != function_offset:
                        bytes_read = struct.unpack("I", table_buffer[read_offset:read_offset+4])[0]
                        read_offset += 4
                except ValueError:
                    delete = True
                    offsets2.pop(offset)
                    continue
                if version == '1.16' and bitness == 64:
                    read_offset += 4
                name_offset = struct.unpack('I', table_buffer[read_offset:read_offset+4])[0]
                function_name = self._readUtf8(function_name_buffer[name_offset:])
                read_offset += 4
                functions[function_offset + start_text] = function_name
        return functions



import logging

LOGGER = logging.getLogger(__name__)

LIEF_AVAILABLE = False
try:
    import lief
    lief.logging.disable()
    LIEF_AVAILABLE = True
except:
    LOGGER.warning("LIEF not available, will not be able to parse data from ELF files.")


def align(v, alignment):
    remainder = v % alignment
    if remainder == 0:
        return v
    else:
        return v + (alignment - remainder)


class ElfFileLoader(object):

    @staticmethod
    def isCompatible(data):
        if not LIEF_AVAILABLE:
            return False
        # check for ELF magic
        return data[:4] == b"\x7FELF"

    @staticmethod
    def getBaseAddress(binary):
        elffile = lief.parse(bytearray(binary))
        # Determine base address of binary
        #
        base_addr = 0
        candidates = [0xFFFFFFFFFFFFFFFF]
        for section in elffile.sections:
            if section.virtual_address:
                candidates.append(section.virtual_address - section.offset)
        if len(candidates) > 1:
            base_addr = min(candidates)
        return base_addr

    @staticmethod
    def mapBinary(binary):
        """
        map the ELF file sections and segments into a contiguous bytearray
        as if into virtual memory with the given base address.
        """
        # ELFFile needs a file-like object...
        # Attention: for Python 2.x use the cStringIO package for StringIO
        elffile = lief.parse(bytearray(binary))
        base_addr = ElfFileLoader.getBaseAddress(binary)

        LOGGER.debug("ELF: base address: 0x%x", base_addr)

        # a segment may contain 0 or more sections.
        # ref: https://stackoverflow.com/a/14382477/87207
        #
        # i'm not sure if a section may be found outside of a segment.
        # therefore, lets load segments first, and then load sections over them.
        # we expect the section data to overwrite the segment data; however,
        # it should be exactly the same data.

        # find min and max virtual addresses.
        max_virtual_address = 0
        min_virtual_address = 0xFFFFFFFFFFFFFFFF
        min_raw_offset = 0xFFFFFFFFFFFFFFFF

        # find begin of the first section/segment and end of the last section/segment.
        for section in sorted(elffile.sections, key=lambda section: section.size, reverse=True):
            if not section.virtual_address:
                continue

            max_virtual_address = max(max_virtual_address, section.size + section.virtual_address)
            min_virtual_address = min(min_virtual_address, section.virtual_address)
            min_raw_offset = min(min_raw_offset, section.file_offset)

        for segment in elffile.segments:
            if not segment.virtual_address:
                continue
            max_virtual_address = max(max_virtual_address, segment.virtual_size + segment.virtual_address)
            min_virtual_address = min(min_virtual_address, segment.virtual_address)
            min_raw_offset = min(min_raw_offset, segment.file_offset)

        if not max_virtual_address:
            LOGGER.debug("ELF: no section or segment data")
            return bytes()

        # create mapped region.
        # offset 0x0 corresponds to the ELF base address
        virtual_size = max_virtual_address - base_addr
        LOGGER.debug("ELF: max virtual section offset: 0x%x", max_virtual_address)
        LOGGER.debug("ELF: mapped size: 0x%x", virtual_size)
        LOGGER.debug("ELF: min raw offset: 0x%x", min_raw_offset)
        mapped_binary = bytearray(align(virtual_size, 0x1000))

        # map segments.
        # segments may contains 0 or more sections,
        # so we do segments first.
        #
        # load sections from largest to smallest,
        # because some segments may overlap.
        #
        # technically, we should only have to load PT_LOAD segments,
        # but we do all of them here.
        for segment in sorted(elffile.segments, key=lambda segment: segment.physical_size, reverse=True):
            if not segment.virtual_address:
                continue
            rva = segment.virtual_address - base_addr
            LOGGER.debug("ELF: mapping segment of 0x%04x bytes at 0x%08x-0x%08x (0x%08x)", segment.physical_size, rva, rva + segment.physical_size, segment.virtual_address)
            assert len(segment.content) == segment.physical_size
            mapped_binary[rva:rva + segment.physical_size] = segment.content

        # map sections.
        # may overwrite some segment data, but we expect the content to be identical.
        for section in sorted(elffile.sections, key=lambda section: section.size, reverse=True):
            if not section.virtual_address:
                continue
            rva = section.virtual_address - base_addr
            LOGGER.debug("ELF: mapping section of 0x%04x bytes at 0x%08x-0x%08x (0x%08x)", section.size, rva, rva + section.size, section.virtual_address)
            assert len(section.content) == section.size
            mapped_binary[rva:rva + section.size] = section.content

        # map header.
        # we consider the headers to be any data found before the first section/segment
        if min_raw_offset != 0:
            LOGGER.debug("ELF: mapping 0x%x bytes of header at 0x0 (0x%x)", min_raw_offset, base_addr)
            mapped_binary[0:min_raw_offset] = binary[0:min_raw_offset]

        LOGGER.debug("ELF: final mapped size: 0x%x", len(mapped_binary))
        return bytes(mapped_binary)

    @staticmethod
    def getBitness(binary):
        # TODO add machine types whenever we add more architectures
        elffile = lief.parse(bytearray(binary))
        machine_type = elffile.header.machine_type
        if machine_type == lief.ELF.ARCH.x86_64:
            return 64
        elif machine_type == lief.ELF.ARCH.i386:
            return 32
        return 0

    @staticmethod
    def mergeCodeAreas(code_areas):
        merged_code_areas = sorted(code_areas)
        result = []
        index = 0
        while index < len(merged_code_areas) - 1:
            this_area = merged_code_areas[index]
            next_area = merged_code_areas[index + 1]
            if this_area[1] != next_area[0]:
                result.append(this_area)
                index += 1
            else:
                merged_code_areas = merged_code_areas[:index] + [[this_area[0], next_area[1]]] + merged_code_areas[index + 2:]
        return merged_code_areas

    @staticmethod
    def getCodeAreas(binary):
        # TODO add machine types whenever we add more architectures
        elffile = lief.parse(bytearray(binary))
        code_areas = []
        for section in elffile.sections:
            # SHF_EXECINSTR = 4
            if section.flags & 0x4:
                section_start = section.virtual_address
                section_size = section.size
                if section_size % section.alignment != 0:
                    section_size += section.alignment - (section_size % section.alignment)
                section_end = section_start + section_size
                code_areas.append([section_start, section_end])              
        return ElfFileLoader.mergeCodeAreas(code_areas)



#!/usr/bin/python

import os
import json
import logging

import lief
lief.logging.disable()

from .AbstractLabelProvider import AbstractLabelProvider
from smda.common.labelprovider.OrdinalHelper import OrdinalHelper

LOGGER = logging.getLogger(__name__)


class WinApiResolver(AbstractLabelProvider):
    """ Minimal WinAPI reference resolver, extracted from ApiScout """

    def __init__(self, config):
        self._config = config
        self._has_64bit = False
        self._api_map = {
            "lief": {}
        }
        self._os_name = None
        self._is_buffer = False
        for os_name, db_filepath in self._config.API_COLLECTION_FILES.items():
            self._loadDbFile(os_name, db_filepath)
            self._os_name = os_name

    def update(self, binary_info):
        self._is_buffer = binary_info.is_buffer
        if not self._is_buffer:
            #setup import table info from LIEF
            lief_binary = lief.parse(bytearray(binary_info.raw_data))
            if not isinstance(lief_binary, lief.PE.Binary):
                return
            for imported_library in lief_binary.imports:
                for func in imported_library.entries:
                    if func.name:
                        self._api_map["lief"][func.iat_address + binary_info.base_addr] = (imported_library.name.lower(), func.name)
                    elif func.is_ordinal:
                        resolved_ordinal = OrdinalHelper.resolveOrdinal(imported_library.name.lower(), func.ordinal)
                        ordinal_name = resolved_ordinal if resolved_ordinal else "#%s" % func.ordinal
                        self._api_map["lief"][func.iat_address + binary_info.base_addr] = (imported_library.name.lower(), ordinal_name)

    def setOsName(self, os_name):
        self._os_name = os_name

    def _loadDbFile(self, os_name, db_filepath):
        api_db = {}
        if os.path.isfile(db_filepath):
            with open(db_filepath, "r") as f_json:
                api_db = json.loads(f_json.read())
        else:
            LOGGER.error("Can't find ApiScout collection file: \"%s\" -- continuing without ApiResolver.", db_filepath)
            return
        num_apis_loaded = 0
        api_map = {}
        for dll_entry in api_db["dlls"]:
            LOGGER.debug("  building address map for: %s", dll_entry)
            for export in api_db["dlls"][dll_entry]["exports"]:
                num_apis_loaded += 1
                api_name = "%s" % (export["name"])
                if api_name == "None":
                    api_name = "None<{}>".format(export["ordinal"])
                dll_name = "_".join(dll_entry.split("_")[2:])
                bitness = api_db["dlls"][dll_entry]["bitness"]
                self._has_64bit |= bitness == 64
                base_address = api_db["dlls"][dll_entry]["base_address"]
                virtual_address = base_address + export["address"]
                api_map[virtual_address] = (dll_name, api_name)
        LOGGER.debug("loaded %d exports from %d DLLs (%s).", num_apis_loaded, len(api_db["dlls"]), api_db["os_name"])
        self._api_map[os_name] = api_map

    def isApiProvider(self):
        """Returns whether the get_api(..) function of the AbstractLabelProvider is functional"""
        return True

    def getApi(self, to_addr, absolute_addr):
        """If the LabelProvider has any information about a used API for the given address, return (dll, api), else return (None, None)"""
        # if we work on a dump, use ApiScout method:
        if self._is_buffer:
            if self._os_name and self._os_name in self._api_map:
                return self._api_map[self._os_name].get(absolute_addr, (None, None))
            else:
                return (None, None)
        # otherwise take import table info from LIEF
        else:
            return self._api_map["lief"].get(to_addr, (None, None))



import hashlib
import datetime
import traceback
import logging

from smda.utility.FileLoader import FileLoader
from smda.utility.MemoryFileLoader import MemoryFileLoader
from smda.SmdaConfig import SmdaConfig
from smda.common.BinaryInfo import BinaryInfo
from smda.common.SmdaReport import SmdaReport
from .intel.IntelDisassembler import IntelDisassembler
from .ida.IdaExporter import IdaExporter

LOGGER = logging.getLogger(__name__)

class Disassembler(object):

    def __init__(self, config=None, backend="intel"):
        if config is None:
            config = SmdaConfig()
        self.config = config
        self.disassembler = None
        if backend == "intel":
            self.disassembler = IntelDisassembler(self.config)
        elif backend == "IDA":
            self.disassembler = IdaExporter(self.config)
        self._start_time = None
        self._timeout = 0
        # cache the last DisassemblyResult
        self.disassembly = None

    def _getDurationInSeconds(self, start_ts, end_ts):
        return (end_ts - start_ts).seconds + ((end_ts - start_ts).microseconds / 1000000.0)

    def _callbackAnalysisTimeout(self):
        if not self._timeout:
            return False
        time_diff = datetime.datetime.utcnow() - self._start_time
        return time_diff.seconds >= self._timeout

    def disassembleFile(self, file_path, pdb_path=""):
        loader = FileLoader(file_path, map_file=True)
        file_content = loader.getData()
        binary_info = BinaryInfo(file_content)
        binary_info.raw_data = loader.getRawData()
        # we want the SHA256 of the unmapped file not how we mapped it to memory
        binary_info.sha256 = hashlib.sha256(binary_info.raw_data).hexdigest()
        binary_info.file_path = file_path
        binary_info.base_addr = loader.getBaseAddress()
        binary_info.bitness = loader.getBitness()
        binary_info.code_areas = loader.getCodeAreas()
        start = datetime.datetime.utcnow()
        try:
            self.disassembler.addPdbFile(binary_info, pdb_path)
            smda_report = self._disassemble(binary_info, timeout=self.config.TIMEOUT)
            if self.config.STORE_BUFFER:
                smda_report.buffer = file_content
        except Exception as exc:
            LOGGER.error("An error occurred while disassembling file.")
            # print("-> an error occured (", str(exc), ").")
            smda_report = self._createErrorReport(start, exc)
        return smda_report

    def disassembleUnmappedBuffer(self, file_content):
        loader = MemoryFileLoader(file_content, map_file=True)
        file_content = loader.getData()
        binary_info = BinaryInfo(file_content)
        binary_info.raw_data = loader.getRawData()
        # we want the SHA256 of the unmapped file not how we mapped it to memory
        binary_info.sha256 = hashlib.sha256(binary_info.raw_data).hexdigest()
        binary_info.file_path = ""
        binary_info.base_addr = loader.getBaseAddress()
        binary_info.bitness = loader.getBitness()
        binary_info.code_areas = loader.getCodeAreas()
        start = datetime.datetime.utcnow()
        try:
            smda_report = self._disassemble(binary_info, timeout=self.config.TIMEOUT)
            if self.config.STORE_BUFFER:
                smda_report.buffer = file_content
        except Exception as exc:
            LOGGER.error("An error occurred while disassembling unmapped buffer.")
            # print("-> an error occured (", str(exc), ").")
            smda_report = self._createErrorReport(start, exc)
        return smda_report

    def disassembleBuffer(self, file_content, base_addr, bitness=None):
        start = datetime.datetime.utcnow()
        try:
            binary_info = BinaryInfo(file_content)
            binary_info.base_addr = base_addr
            binary_info.bitness = bitness
            binary_info.is_buffer = True
            smda_report = self._disassemble(binary_info, timeout=self.config.TIMEOUT)
            if self.config.STORE_BUFFER:
                smda_report.buffer = file_content
        except Exception as exc:
            LOGGER.error("An error occurred while disassembling buffer.")
            # print("-> an error occured (", str(exc), ").")
            smda_report = self._createErrorReport(start, exc)
        return smda_report

    def _disassemble(self, binary_info, timeout=0):
        self._start_time = datetime.datetime.utcnow()
        self._timeout = timeout
        self.disassembly = self.disassembler.analyzeBuffer(binary_info, self._callbackAnalysisTimeout)
        return SmdaReport(self.disassembly, config=self.config)

    def _createErrorReport(self, start, exception):
        report = SmdaReport(config=self.config)
        report.smda_version = self.config.VERSION
        report.status = "error"
        report.execution_time = self._getDurationInSeconds(start, datetime.datetime.utcnow())
        report.message = traceback.format_exc()
        return report



import datetime
import struct
import sys

from smda.common.BasicBlock import BasicBlock


class DisassemblyResult(object):

    def __init__(self):
        self.analysis_start_ts = datetime.datetime.utcnow()
        self.analysis_end_ts = self.analysis_start_ts
        self.analysis_timeout = False
        self.binary_info = None
        self.identified_alignment = 0
        self.oep = None
        self.code_map = {}
        self.data_map = set([])
        # key: offset, value: {"type": <str>, "instruction_bytes": <hexstr>}
        self.errors = {}
        # stored as key:
        self.functions = {}
        self.recursive_functions = set([])
        self.leaf_functions = set([])
        self.thunk_functions = set([])
        self.exported_functions = set([])
        self.failed_analysis_addr = []
        self.function_borders = {}
        # stored as key: int(i.address) = (i.size, i.mnemonic, i.op_str)
        self.instructions = {}
        self.ins2fn = {}
        self.language = {}
        self.data_refs_from = {}
        self.data_refs_to = {}
        self.code_refs_from = {}
        self.code_refs_to = {}
        # key: address of API in target DLL, value: {referencing_addr, api_name, dll_name}
        self.apis = {}
        self.addr_to_api = {}
        # address:name
        self.function_symbols = {}
        self.candidates = {}
        self._confidence_threshold = 0.0
        self.code_areas = []
        self.smda_version = ""

    def setBinaryInfo(self, binary_info):
        self.binary_info = binary_info
        exported = binary_info.getExportedFunctions()
        if exported is not None:
            self.exported_functions = set([key + binary_info.base_addr for key in exported.keys()])
        self.oep = binary_info.getOep()

    def getByte(self, addr):
        if self.isAddrWithinMemoryImage(addr):
            return self.binary_info.binary[addr - self.binary_info.base_addr]
        return None

    def getRawByte(self, offset):
        return self.binary_info.binary[offset]

    def getBytes(self, addr, num_bytes):
        if self.isAddrWithinMemoryImage(addr):
            rel_start_addr = addr - self.binary_info.base_addr
            return self.binary_info.binary[rel_start_addr:rel_start_addr + num_bytes]
        return None

    def getRawBytes(self, offset, num_bytes):
        return self.binary_info.binary[offset:offset + num_bytes]

    def setConfidenceThreshold(self, threshold):
        self._confidence_threshold = threshold

    def getConfidenceThreshold(self):
        return self._confidence_threshold

    def getAnalysisDuration(self):
        return (self.analysis_end_ts - self.analysis_start_ts).seconds + ((self.analysis_end_ts - self.analysis_start_ts).microseconds / 1000000.0)

    def getAnalysisOutcome(self):
        outcome = "ok"
        if self.analysis_timeout:
            outcome = "timeout"
        return outcome

    def getFunctions(self):
        return sorted(self.functions.keys())

    def getBlocks(self, function_addr):
        disasm_blocks = []
        if function_addr in self.functions:
            disasm_blocks = self.functions[function_addr]
        bblocks = []

        for block in disasm_blocks:
            bblock = BasicBlock()
            bblock.start_addr = block[0][0]
            bblock.end_addr = block[-1][0]
            bblock.instructions = [ins[0] for ins in block]
            if bblock.end_addr in self.code_refs_from:
                bblock.successors = [ref for ref in self.code_refs_from[bblock.end_addr]]
            bblocks.append(bblock)
        return bblocks

    def _transformInstruction(self, ins_tuple):
        ins_addr, _, ins_mnem, ins_ops, ins_raw_bytes = ins_tuple
        # python3  and python2 do handling differently...
        if sys.version_info >= (3, 0):
            ins_hexbytes = "".join(["%02x" % c for c in ins_tuple[4]])
        else:
            ins_hexbytes = ins_raw_bytes.encode("hex")
        return [ins_addr, ins_hexbytes, str(ins_mnem), str(ins_ops)]

    def getBlocksAsDict(self, function_addr):
        blocks = {}
        for block in self.functions[function_addr]:
            instructions = []
            for ins in block:
                instructions.append(self._transformInstruction(ins))
            blocks[instructions[0][0]] = instructions
        return blocks

    def getInstructions(self, block):
        return block.instructions

    def getMnemonic(self, instruction_addr):
        if instruction_addr in self.instructions:
            return self.instructions[instruction_addr][0]
        return ""

    def isCode(self, addr):
        return addr in self.code_map

    def isAddrWithinMemoryImage(self, destination):
        if destination is not None:
            return self.binary_info.base_addr <= destination < (self.binary_info.base_addr + self.binary_info.binary_size)
        return False

    def dereferenceDword(self, addr):
        if self.isAddrWithinMemoryImage(addr):
            rel_start_addr = addr - self.binary_info.base_addr
            rel_end_addr = rel_start_addr + 4
            extracted_dword = self.binary_info.binary[rel_start_addr:rel_end_addr]
            return struct.unpack("I", extracted_dword)[0]
        return None

    def dereferenceQword(self, addr):
        if self.isAddrWithinMemoryImage(addr):
            rel_start_addr = addr - self.binary_info.base_addr
            rel_end_addr = rel_start_addr + 8
            extracted_qword = self.binary_info.binary[rel_start_addr:rel_end_addr]
            return struct.unpack("Q", extracted_qword)[0]
        return None

    def addCodeRefs(self, addr_from, addr_to):
        refs_from = self.code_refs_from.get(addr_from, set([]))
        refs_from.update([addr_to])
        self.code_refs_from[addr_from] = refs_from
        refs_to = self.code_refs_to.get(addr_to, set([]))
        refs_to.update([addr_from])
        self.code_refs_to[addr_to] = refs_to

    def removeCodeRefs(self, addr_from, addr_to):
        refs_from = self.code_refs_from.get(addr_from, set([]))
        refs_from.discard(addr_to)
        self.code_refs_from[addr_from] = refs_from
        refs_to = self.code_refs_to.get(addr_to, set([]))
        refs_to.discard(addr_from)
        self.code_refs_to[addr_to] = refs_to

    def addDataRefs(self, addr_from, addr_to):
        refs_from = self.data_refs_from.get(addr_from, set([]))
        refs_from.update([addr_to])
        self.data_refs_from[addr_from] = refs_from
        refs_to = self.data_refs_to.get(addr_to, set([]))
        refs_to.update([addr_from])
        self.data_refs_to[addr_to] = refs_to

    def removeDataRefs(self, addr_from, addr_to):
        refs_from = self.data_refs_from.get(addr_from, set([]))
        refs_from.discard(addr_to)
        self.data_refs_from[addr_from] = refs_from
        refs_to = self.data_refs_to.get(addr_to, set([]))
        refs_to.discard(addr_from)
        self.data_refs_to[addr_to] = refs_to

    def getBlockRefs(self, func_addr):
        """ blocks refs should stay within function context, thus kill all references outside function """
        block_refs = {}
        ins_addrs = set([])
        for block in self.functions[func_addr]:
            for ins in block:
                ins_addr = ins[0]
                ins_addrs.add(ins_addr)
        for block in self.functions[func_addr]:
            last_ins_addr = block[-1][0]
            if last_ins_addr in self.code_refs_from:
                verified_refs = sorted(list(ins_addrs.intersection(self.code_refs_from[last_ins_addr])))
                if verified_refs:
                    block_refs[block[0][0]] = verified_refs
        return block_refs

    def getInRefs(self, func_addr):
        in_refs = []
        if func_addr in self.code_refs_to:
            in_refs = list(self.code_refs_to[func_addr])
        return sorted(in_refs)

    def getOutRefs(self, func_addr):
        ins_addrs = set([])
        code_refs = []
        out_refs = {}
        for block in self.functions[func_addr]:
            for ins in block:
                ins_addr = ins[0]
                ins_addrs.add(ins_addr)
                if ins_addr in self.code_refs_from:
                    for to_addr in self.code_refs_from[ins_addr]:
                        code_refs.append([ins_addr, to_addr])
        # function may be recursive
        if func_addr in ins_addrs:
            ins_addrs.remove(func_addr)
        # reduce outrefs to addresses within the memory image
        max_addr = self.binary_info.base_addr + self.binary_info.binary_size
        image_refs = [ref for ref in code_refs if self.binary_info.base_addr <= ref[1] <= max_addr]
        for ref in image_refs:
            if ref[1] in ins_addrs:
                continue
            if ref[0] not in out_refs:
                out_refs[ref[0]] = []
            out_refs[ref[0]].append(ref[1])
        return {src: sorted(dst) for src, dst in out_refs.items()}

    def isRecursiveFunction(self, func_addr):
        ins_addrs = set([])
        out_refs = set([])
        for block in self.functions[func_addr]:
            for ins in block:
                ins_addr = ins[0]
                ins_addrs.add(ins_addr)
                if ins_addr in self.code_refs_from:
                    for to_addr in self.code_refs_from[ins_addr]:
                        out_refs.add(to_addr)
        return func_addr in out_refs

    def isLeafFunction(self, func_addr):
        ins_addrs = set([])
        out_refs = set([])
        for block in self.functions[func_addr]:
            for ins in block:
                ins_addr = ins[0]
                ins_addrs.add(ins_addr)
                if ins_addr in self.code_refs_from:
                    for to_addr in self.code_refs_from[ins_addr]:
                        out_refs.add(to_addr)
        return len(out_refs.difference(ins_addrs)) == 0

    def _initApiRefs(self):
        for api_offset in self.apis:
            api = self.apis[api_offset]
            for ref in api["referencing_addr"]:
                self.addr_to_api[ref] = "%s!%s" % (api["dll_name"], api["api_name"])

    def getAllApiRefs(self):
        all_api_refs = {}
        for function_addr in self.functions:
            all_api_refs.update(self.getApiRefs(function_addr))
        return all_api_refs

    def getApiRefs(self, func_addr):
        if not self.addr_to_api:
            self._initApiRefs()
        api_refs = {}
        for block in self.functions[func_addr]:
            for ins in block:
                if ins[0] in self.addr_to_api:
                    api_refs[ins[0]] = self.addr_to_api[ins[0]]
        return api_refs

    def __str__(self):
        return "-> {:5.2f}s | {:5d} Func (status: {})".format(self.getAnalysisDuration(), len(self.functions), self.getAnalysisOutcome())



import binascii
import datetime
import json
import os
import zipfile

try:
    from StringIO import StringIO ## for Python 2
except ImportError:
    from io import StringIO ## for Python 3

from capstone import Cs, CS_ARCH_X86, CS_MODE_32, CS_MODE_64

from .BinaryInfo import BinaryInfo
from .SmdaFunction import SmdaFunction
from smda.common.CodeXref import CodeXref
from smda.DisassemblyStatistics import DisassemblyStatistics


class SmdaReport(object):

    architecture = None
    base_addr = None
    binary_size = None
    binweight = None
    bitness = None
    buffer = None
    code_areas = None
    code_sections = None
    component = None
    confidence_threshold = None
    disassembly_errors = None
    execution_time = None
    family = None
    filename = None
    identified_alignment = None
    is_library = None
    is_buffer = None
    message = None
    oep = None
    sha256 = None
    smda_version = None
    statistics = None
    # status can be "ok", "timeout", "error"
    status = None
    timestamp = None
    version = None
    xcfg = None

    # on first usage, initialize codexrefs objects for all functions based on inrefs/outrefs (requires knowledge about all functions)
    _has_codexrefs = False
    # in case we need to re-disassemble with more detail, we hold an initialized capstone instance as singleton
    capstone = None

    def __init__(self, disassembly=None, config=None, buffer=None):
        if disassembly is not None:
            self.architecture = disassembly.binary_info.architecture
            self.base_addr = disassembly.binary_info.base_addr
            self.binary_size = disassembly.binary_info.binary_size
            self.binweight = 0
            self.bitness = disassembly.binary_info.bitness
            self.buffer = buffer
            self.code_areas = disassembly.binary_info.code_areas
            self.code_sections = [section for section in disassembly.binary_info.getSections()]
            if self.code_sections is None:
                self.code_sections = []
            self.component = disassembly.binary_info.component
            self.confidence_threshold = disassembly.getConfidenceThreshold()
            self.disassembly_errors = disassembly.errors
            self.execution_time = disassembly.getAnalysisDuration()
            self.family = disassembly.binary_info.family
            self.filename = os.path.basename(disassembly.binary_info.file_path)
            self.identified_alignment = disassembly.identified_alignment
            self.is_library = disassembly.binary_info.is_library
            self.is_buffer = disassembly.binary_info.is_buffer
            self.message = "Analysis finished regularly."
            self.oep = disassembly.binary_info.getOep()
            self.sha256 = disassembly.binary_info.sha256
            self.smda_version = disassembly.smda_version
            self.statistics = DisassemblyStatistics(disassembly)
            self.status = disassembly.getAnalysisOutcome()
            if self.status == "timeout":
                self.message = "Analysis was stopped when running into the timeout."
            self.timestamp = datetime.datetime.utcnow()
            self.version = disassembly.binary_info.version
            self.xcfg = self._convertCfg(disassembly, config=config)

    def _convertCfg(self, disassembly, config=None):
        function_results = {}
        for function_offset in disassembly.functions:
            if self.confidence_threshold and function_offset in disassembly.candidates and disassembly.candidates[function_offset].getConfidence() < self.confidence_threshold:
                continue
            smda_function = SmdaFunction(disassembly, function_offset, config=config, smda_report=self)
            function_results[function_offset] = smda_function
            self.binweight += smda_function.binweight
        return function_results

    @property
    def num_functions(self):
        return len(self.xcfg)

    @property
    def num_blocks(self):
        sum_blocks = 0
        for function in self.getFunctions():
            sum_blocks += function.num_blocks
        return sum_blocks

    @property
    def num_instructions(self):
        sum_instructions = 0
        for function in self.getFunctions():
            sum_instructions += function.num_instructions
        return sum_instructions

    def getBuffer(self):
        return self.buffer

    def getFunction(self, function_addr):
        return self.xcfg[function_addr] if function_addr in self.xcfg else None

    def getFunctions(self):
        for _, smda_function in sorted(self.xcfg.items()):
            yield smda_function

    def getExportedFunctions(self):
        for _, smda_function in sorted(self.xcfg.items()):
            if smda_function.isExported():
                yield smda_function

    def getCapstone(self):
        if self.capstone is None:
            self.capstone = Cs(CS_ARCH_X86, CS_MODE_64) if self.bitness == 64 else Cs(CS_ARCH_X86, CS_MODE_32)
            self.capstone.detail = True
        return self.capstone

    def getSection(self, offset):
        for section in self.code_sections:
            if section[1] <= offset < section[2]:
                return section

    def isAddrWithinMemoryImage(self, offset):
        return self.base_addr <= offset < self.base_addr + self.binary_size

    def initCodeXrefs(self):
        if not self._has_codexrefs:
            # create ins2fn map, and offset to SmdaInstruction map
            ins2fn = {}
            offset2ins = {}
            tmp_functions = []
            for function in self.getFunctions():
                tmp_functions.append(function.offset)
                for instruction in function.getInstructions():
                    ins2fn[instruction.offset] = function
                    offset2ins[instruction.offset] = instruction
            # for all functions, populate code references
            for function in self.getFunctions():
                function.code_inrefs = []
                for inref in function.inrefs:
                    function.code_inrefs.append(CodeXref(offset2ins[inref], offset2ins[function.offset]))
                function.code_outrefs = []
                for outref_src, outref_dsts in function.outrefs.items():
                    for target in outref_dsts:
                        if target in offset2ins:
                            function.code_outrefs.append(CodeXref(offset2ins[outref_src], offset2ins[target]))
            self._has_codexrefs = True

    def _packBuffer(self, buffer):
        # TODO
        # create zip
        # XOR with some key
        # base64
        return b""

    def _unpackBuffer(self, buffer):
        # TODO
        # de-base64
        # XOR with some key
        # read from zip
        return b""

    @classmethod
    def fromFile(cls, file_path):
        smda_json = {}
        if os.path.isfile(file_path):
            with open(file_path, "r") as fjson:
                smda_json = json.load(fjson)
        return SmdaReport.fromDict(smda_json)

    @classmethod
    def fromDict(cls, report_dict):
        smda_report = cls(None)
        smda_report.architecture = report_dict["architecture"]
        smda_report.base_addr = report_dict["base_addr"]
        smda_report.binary_size = report_dict["binary_size"]
        smda_report.bitness = report_dict["bitness"]
        smda_report.code_areas = report_dict["code_areas"]
        smda_report.code_sections = [("", section[1], section[2]) for section in report_dict["code_sections"]] if "code_sections" in report_dict else []
        smda_report.confidence_threshold = report_dict["confidence_threshold"]
        smda_report.disassembly_errors = report_dict["disassembly_errors"]
        smda_report.execution_time = report_dict["execution_time"]
        smda_report.identified_alignment = report_dict["identified_alignment"]
        if "metadata" in report_dict:
            if "binweight" in report_dict["metadata"]:
                smda_report.binweight = report_dict["metadata"]["binweight"]
            if "component" in report_dict["metadata"]:
                smda_report.component = report_dict["metadata"]["component"]
            if "family" in report_dict["metadata"]:
                smda_report.family = report_dict["metadata"]["family"]
            if "filename" in report_dict["metadata"]:
                smda_report.filename = report_dict["metadata"]["filename"]
            if "is_library" in report_dict["metadata"]:
                smda_report.is_library = report_dict["metadata"]["is_library"]
            if "version" in report_dict["metadata"]:
                smda_report.version = report_dict["metadata"]["version"]
            smda_report.is_buffer = report_dict["metadata"]["is_buffer"] if "is_buffer" in report_dict["metadata"] else False
        smda_report.message = report_dict["message"]
        smda_report.oep = report_dict["oep"] if "oep" in report_dict else None
        smda_report.sha256 = report_dict["sha256"]
        smda_report.smda_version = report_dict["smda_version"]
        smda_report.statistics = DisassemblyStatistics.fromDict(report_dict["statistics"])
        smda_report.status = report_dict["status"]
        smda_report.timestamp = datetime.datetime.strptime(report_dict["timestamp"], "%Y-%m-%dT%H-%M-%S")
        binary_info = BinaryInfo(b"")
        binary_info.architecture = smda_report.architecture
        binary_info.base_addr = smda_report.base_addr
        binary_info.binary_size = smda_report.binary_size
        binary_info.oep = smda_report.oep
        smda_report.xcfg = {int(function_addr): SmdaFunction.fromDict(function_dict, binary_info=binary_info, version=smda_report.smda_version, smda_report=smda_report) for function_addr, function_dict in report_dict["xcfg"].items()}
        return smda_report

    def toDict(self):
        return {
            "architecture": self.architecture,
            "base_addr": self.base_addr,
            "binary_size": self.binary_size,
            "bitness": self.bitness,
            "code_areas": self.code_areas,
            "code_sections": [("", section[1], section[2]) for section in self.code_sections],
            "confidence_threshold": self.confidence_threshold,
            "disassembly_errors": self.disassembly_errors,
            "execution_time": self.execution_time,
            "identified_alignment": self.identified_alignment,
            "metadata" : {
                "binweight": self.binweight,
                "component": self.component,
                "family": self.family,
                "filename": self.filename,
                "is_library": self.is_library,
                "is_buffer": self.is_buffer,
                "version": self.version,
            },
            "message": self.message,
            "oep": self.oep,
            "sha256": self.sha256,
            "smda_version": self.smda_version,
            "statistics": self.statistics.toDict(),
            "status": self.status,
            "timestamp": self.timestamp.strftime("%Y-%m-%dT%H-%M-%S"),
            "xcfg": {function_addr: smda_function.toDict() for function_addr, smda_function in self.xcfg.items()}
        }

    def __str__(self):
        if self.status == "error":
            return "{:>6.3f}s -> {}".format(self.execution_time, self.message)
        return "{:>6.3f}s -> (architecture: {}.{}bit, base_addr: 0x{:08x}): {} functions".format(self.execution_time, self.architecture, self.bitness, self.base_addr, len(self.xcfg))



#!/usr/bin/python

import logging

from smda.utility.PeFileLoader import PeFileLoader
from .AbstractLabelProvider import AbstractLabelProvider

LOGGER = logging.getLogger(__name__)

try:
    import pdbparse
    from pdbparse.undname import undname
except:
    pdbparse = None
    LOGGER.debug("3rd party library pdbparse (use fork @ https://github.com/VPaulV/pdbparse) not installed - won't be able to extract symbols from PDB files where available.")


class DummyOmap(object):
    def remap(self, addr):
        return addr


class PdbSymbolProvider(AbstractLabelProvider):
    """ Minimal resolver for PDB symbols """

    def __init__(self, config):
        self._config = config
        self._base_addr = 0
        # addr:func_name
        self._func_symbols = {}

    def isSymbolProvider(self):
        return True

    def _parseOep(self, data):
        oep_rva = PeFileLoader.getOEP(data)
        if oep_rva:
            self._func_symbols[self._base_addr + oep_rva] = "original_entry_point"

    def update(self, binary_info):
        self._base_addr = binary_info.base_addr
        if not binary_info.file_path:
            return
        data = ""
        with open(binary_info.file_path, "rb") as fin:
            data = fin.read(16)
        self._parseOep(data)
        if data[:15] != b"Microsoft C/C++" or pdbparse is None:
            return
        try:
            pdb = pdbparse.parse(binary_info.file_path)
            self._parseSymbols(pdb)
        except Exception as exc:
            LOGGER.error("Failed parsing \"%s\" with exception type: %s", binary_info.file_path, type(exc))

    def _parseSymbols(self, pdb):
        try:
            sects = pdb.STREAM_SECT_HDR_ORIG.sections
            omap = pdb.STREAM_OMAP_FROM_SRC
        except AttributeError:
            sects = pdb.STREAM_SECT_HDR.sections
            omap = DummyOmap()
        gsyms = pdb.STREAM_GSYM
        for sym in gsyms.globals:
            try:
                off = sym.offset
                if len(sects) < sym.segment:
                    continue
                virt_base = sects[sym.segment - 1].VirtualAddress
                function_address = (self._base_addr + omap.remap(off + virt_base))
                demangled_name = undname(sym.name)
                if sym.symtype == 2:
                    # print("0x%x + 0x%x + 0x%x = 0x%x: %s || %s (type: %d)" % (self._base_addr, off, virt_base, function_address, sym.name, demangled_name, sym.symtype))
                    self._func_symbols[function_address] = demangled_name
            except AttributeError:
                pass

    def getSymbol(self, address):
        return self._func_symbols.get(address, "")

    def getFunctionSymbols(self):
        return self._func_symbols



#!/usr/bin/python
import lief
lief.logging.disable()

from .AbstractLabelProvider import AbstractLabelProvider


class ElfApiResolver(AbstractLabelProvider):
    """ Minimal ELF API reference resolver, extracting APIs from ELF imports """

    def __init__(self, config):
        self._api_map = {
            "lief": {}
        }

    def update(self, binary_info):
        if binary_info.is_buffer:
            # cannot reconstruct from shellcode/memory dump at this time
            return

        else:
            lief_binary = lief.parse(bytearray(binary_info.raw_data))

            if not isinstance(lief_binary, lief.ELF.Binary):
                return

            for relocation in lief_binary.relocations:
                if not relocation.has_symbol:
                    # doesn't have a name, we won't care about it
                    continue
                if not relocation.symbol.imported:
                    # only interested in APIs from external sources
                    continue
                if not relocation.symbol.is_function:
                    # only interested in APIs (which are functions)
                    continue

                # we can't really say what library the symbol came from
                # however, we can treat the version (if present) as relevant metadata?
                # note: this only works for GNU binaries, such as for Linux
                lib = None
                if relocation.symbol.has_version and relocation.symbol.symbol_version.has_auxiliary_version:
                    # like "GLIBC_2.2.5"
                    lib = relocation.symbol.symbol_version.symbol_version_auxiliary.name

                name = relocation.symbol.name
                address = relocation.address

                self._api_map["lief"][address] = (lib, name)

    def isApiProvider(self):
        """Returns whether the get_api(..) function of the AbstractLabelProvider is functional"""
        return True

    def getApi(self, to_addr, absolute_addr):
        """
        If the LabelProvider has any information about a used API for the given address, return (dll, api), else return (None, None).

        May return None for the `dll` if it cannot be determined.
        When it can be determined for ELF files, the `dll` field should be interpreted as the API version rather than shared library name.
        For example: "GLIBC_2.2.5".
        """
        return self._api_map["lief"].get(to_addr, (None, None))



import re
import struct
import logging
from collections import Counter

from .definitions import COMMON_START_BYTES

LOGGER = logging.getLogger(__name__)


class BitnessAnalyzer(object):

    def determineBitnessFromFile(self, filepath):
        LOGGER.debug("Running Bitness test on %s", filepath)
        with open(filepath, "rb") as infile:
            if re.search(r"[0-9a-fA-F]{64}_dump_0x[0-9a-fA-F]+$", filepath):
                return self.determineBitness(binary=infile.read())
        return 0

    def determineBitnessFromDisassembly(self, disassembly):
        LOGGER.debug("Running Bitness test on binary data of DisassemblyResult")
        return self.determineBitness(binary=disassembly.binary_info.binary)

    def determineBitness(self, binary):
        candidate_first_bytes = {"32": Counter(), "64": Counter()}
        # check for potential call instructions and collect their first bytes
        for bitness in ["32", "64"]:
            for call_match in re.finditer(b"\xE8", binary):
                if len(binary) - call_match.start() > 5:
                    packed_call = binary[call_match.start() + 1:call_match.start() + 5]
                    rel_call_offset = struct.unpack("i", packed_call)[0]
                    call_destination = (rel_call_offset + call_match.start() + 5)  # & bitmask
                    if call_destination > 0 and call_destination < len(binary):
                        first_byte = binary[call_destination]
                        candidate_first_bytes[bitness][first_byte] += 1
        score = {"32": 0, "64": 0}
        for bitness in ["32", "64"]:
            for candidate_sequence in candidate_first_bytes[bitness]:
                if isinstance(candidate_sequence, int):
                    candidate_sequence = "%02x" % candidate_sequence
                elif isinstance(candidate_sequence, str):
                    candidate_sequence = candidate_sequence.encode("hex")
                for common_sequence, sequence_score in COMMON_START_BYTES[bitness].items():
                    if candidate_sequence == str(common_sequence):
                        score[bitness] += sequence_score * 1.0
        total_score = max(score["32"] + score["64"], 1)
        score["32"] /= total_score
        score["64"] /= total_score
        LOGGER.debug("Bitness scores: %5.2f (32bit), %5.2f (64bit)", score["32"], score["64"])
        return 64 if score["32"] < score["64"] else 32



import struct
import re
import logging

LOGGER = logging.getLogger(__name__)


class IndirectCallAnalyzer(object):
    """ Perform basic dataflow analysis to resolve indirect call targets """

    def __init__(self, disassembler):
        self.disassembler = disassembler
        self.disassembly = self.disassembler.disassembly
        self.current_calling_addr = 0
        self.state = None

    def searchBlock(self, analysis_state, address):
        for block in analysis_state.getBlocks():
            if address in [i[0] for i in block]:
                return block
        return []

    def getDword(self, addr):
        if not self.disassembly.isAddrWithinMemoryImage(addr):
            return None
        return struct.unpack("I", self.disassembly.getBytes(addr, 4))[0]

    def processBlock(self, analysis_state, block, registers, register_name, processed, depth):
        if not block:
            return False
        if block in processed:
            LOGGER.debug("already processed block 0x%08x; skipping", block[0][0])
            return False
        processed.append(block)
        LOGGER.debug("start processing block: 0x%08x\nlooking for register %s", block[0][0], register_name)
        abs_value_found = False
        for ins in reversed(block):
            LOGGER.debug("0x%08x: %s %s", ins[0], ins[2], ins[3])
            if ins[2] == "mov":
                #mov <reg>, <reg>
                match1 = re.match(r"(?P<reg1>[a-z]{3}), (?P<reg2>[a-z]{3})$", ins[3])
                if match1:
                    if match1.group("reg1") == register_name:
                        register_name = match1.group("reg2")
                #mov <reg>, <const>
                match2 = re.match(r"(?P<reg>[a-z]{3}), (?P<val>0x[0-9a-f]{,8})$", ins[3])
                if match2:
                    registers[match2.group("reg")] = int(match2.group("val"), 16)
                    LOGGER.debug("**moved value 0x%08x to register %s", int(match2.group("val"), 16), match2.group("reg"))
                    if match2.group("reg") == register_name:
                        abs_value_found = True
                #mov <reg>, dword ptr [<addr>]
                match3 = re.match(r"(?P<reg>[a-z]{3}), dword ptr \[(?P<addr>0x[0-9a-f]{,8})\]$", ins[3])
                if match3:
                    # HACK: test to see if the address points to a import and
                    # use that instead of the actual memory value
                    addr = int(match3.group("addr"), 16)
                    dll, api = self.disassembler.resolveApi(addr, addr)
                    if dll or api:
                        registers[match3.group("reg")] = addr
                        LOGGER.debug("**moved API ref (%s:%s) @0x%08x to register %s", dll, api, addr, match3.group("reg"))
                        if match3.group("reg") == register_name:
                            abs_value_found = True
                    else:
                        dword = self.getDword(addr)
                        if dword:
                            registers[match3.group("reg")] = dword
                            LOGGER.debug("**moved value 0x%08x to register %s", dword, match3.group("reg"))
                            if match3.group("reg") == register_name:
                                abs_value_found = True
                #mov <reg>, qword ptr [reg + <addr>]
                match4 = re.match(r"(?P<reg>[a-z]{3}), qword ptr \[rip \+ (?P<addr>0x[0-9a-f]{,8})\]$", ins[3])
                if match4:
                    rip = ins[0] + ins[1]
                    dword = self.getDword(rip + int(match4.group("addr"), 16))
                    if dword:
                        registers[match4.group("reg")] = rip + dword
                        LOGGER.debug("**moved value 0x%08x + 0x%08x == 0x%08x to register %s", rip, dword, rip + dword, match4.group("reg"))
                        if match4.group("reg") == register_name:
                            abs_value_found = True
            elif ins[2] == "lea":
                #lea <reg>, dword ptr [<addr>]
                match1 = re.match(r"(?P<reg>[a-z]{3}), dword ptr \[(?P<addr>0x[0-9a-f]{,8})\]$", ins[3])
                if match1:
                    dword = self.getDword(int(match1.group("addr"), 16))
                    if dword:
                        registers[match1.group("reg")] = dword
                        LOGGER.debug("**moved value 0x%08x to register %s", dword, match1.group("reg"))
                        if match1.group("reg") == register_name:
                            abs_value_found = True
                # not handled: lea <reg>, dword ptr [<reg> +- <val>]
                # requires state-keeping of multiple registers
            # there exist potentially many more way how the register being called can be calculated
            # for now we ignore them
            elif ins[2] == "other instruction":
                pass
            #if the absolute value was found for the call <reg> instruction, detect API
            if abs_value_found:
                candidate = registers[register_name] if register_name in registers else None
                self.state.setLeaf(False)
                if candidate:
                    LOGGER.debug("candidate: 0x%x - %s, register: %s", candidate, ins[3], register_name)
                    dll, api = self.disassembler.resolveApi(candidate, candidate)
                    if dll or api:
                        LOGGER.debug("successfully resolved: %s %s", dll, api)
                        api_entry = {"referencing_addr": [], "dll_name": dll, "api_name": api}
                        if candidate in self.disassembly.apis:
                            api_entry = self.disassembly.apis[candidate]
                        if self.current_calling_addr not in api_entry["referencing_addr"]:
                            api_entry["referencing_addr"].append(self.current_calling_addr)
                        self.disassembly.apis[candidate] = api_entry
                    elif self.disassembly.isAddrWithinMemoryImage(candidate):
                        LOGGER.debug("successfully resolved: 0x%x", candidate)
                        self.disassembler.fc_manager.addCandidate(candidate, reference_source=self.current_calling_addr)
                    else:
                        LOGGER.debug("candidate not resolved")
                else:
                    LOGGER.debug("no candidate to resolved")

                return True
        #process previous blocks
        if depth >= 0:
            refs_in = [
                fr for (fr, to) in analysis_state.code_refs
                if to == block[0][0] and
                fr not in [ins[0] for block in processed for ins in block]
            ]
            LOGGER.debug("start processing previous blocks")
            if any(self.processBlock(analysis_state, b, registers.copy(), register_name, processed, depth - 1) for b in [self.searchBlock(analysis_state, i) for i in refs_in]):
                return True

    def resolveRegisterCalls(self, analysis_state, block_depth=3):
        # after block reconstruction do simple data flow analysis to resolve open cases like "call <register>" as stored in self.call_register_ins
        for calling_addr in analysis_state.call_register_ins:
            LOGGER.debug("#" * 20)
            self.current_calling_addr = calling_addr
            self.state = analysis_state
            start_block = [ins for ins in self.searchBlock(analysis_state, calling_addr) if ins[0] <= calling_addr]
            if start_block:
                self.processBlock(analysis_state, start_block, dict(), start_block[-1][3], list(), block_depth)



#!/usr/bin/python

import datetime
import logging
import re

from capstone import Cs, CS_ARCH_X86, CS_MODE_32, CS_MODE_64

from smda.DisassemblyResult import DisassemblyResult
from smda.common.BinaryInfo import BinaryInfo
from smda.common.labelprovider.WinApiResolver import WinApiResolver
from smda.common.labelprovider.ElfSymbolProvider import ElfSymbolProvider
from smda.common.labelprovider.ElfApiResolver import ElfApiResolver
from smda.common.labelprovider.PdbSymbolProvider import PdbSymbolProvider
from smda.common.labelprovider.GoLabelProvider import GoSymbolProvider
from smda.common.TailcallAnalyzer import TailcallAnalyzer
from .definitions import CJMP_INS, LOOP_INS, JMP_INS, CALL_INS, RET_INS, REGS_32BIT, REGS_64BIT, DOUBLE_ZERO
from .FunctionCandidateManager import FunctionCandidateManager
from .FunctionAnalysisState import FunctionAnalysisState
from .IndirectCallAnalyzer import IndirectCallAnalyzer
from .JumpTableAnalyzer import JumpTableAnalyzer
from .MnemonicTfIdf import MnemonicTfIdf
from .BitnessAnalyzer import BitnessAnalyzer

LOGGER = logging.getLogger(__name__)


class SimpleIns(object):
    address = None
    size = None
    mnemonic = None
    op_str = None
    bytes = None

    def __init__(self, address, size, mnemonic, op_str, bytes):
        self.address = address
        self.size = size
        self.mnemonic = mnemonic
        self.op_str = op_str
        self.bytes = bytes


class IntelDisassembler(object):

    def __init__(self, config, forced_bitness=None):
        self.config = config
        self._forced_bitness = forced_bitness
        self.capstone = None
        self._tfidf = None
        self.binary_info = None
        self.label_providers = []
        self._addLabelProviders()
        self.fc_manager = None
        self.tailcall_analyzer = None
        self.indcall_analyzer = None
        self.jumptable_analyzer = None
        self.disassembly = DisassemblyResult()
        self.disassembly.smda_version = config.VERSION
        self.disassembly.setConfidenceThreshold(config.CONFIDENCE_THRESHOLD)

    def _initCapstone(self):
        self.capstone = Cs(CS_ARCH_X86, CS_MODE_64) if self.disassembly.binary_info.bitness == 64 else Cs(CS_ARCH_X86, CS_MODE_32)

    def _initTfIdf(self):
        self._tfidf = MnemonicTfIdf(bitness=64) if self.disassembly.binary_info.bitness == 64 else MnemonicTfIdf(bitness=32)

    def getBitMask(self):
        if self.disassembly.binary_info.bitness == 64:
            return 0xFFFFFFFFFFFFFFFF
        return 0xFFFFFFFF

    def _addLabelProviders(self):
        self.label_providers.append(WinApiResolver(self.config))
        self.label_providers.append(ElfApiResolver(self.config))
        self.label_providers.append(ElfSymbolProvider(self.config))
        self.label_providers.append(PdbSymbolProvider(self.config))
        self.label_providers.append(GoSymbolProvider(self.config))

    def _updateLabelProviders(self, binary_info):
        for provider in self.label_providers:
            provider.update(binary_info)

    def addPdbFile(self, binary_info, pdb_path):
        LOGGER.debug("adding PDB file: %s", pdb_path)
        if pdb_path and binary_info.base_addr:
            pdb_info = BinaryInfo(b"")
            pdb_info.file_path = pdb_path
            pdb_info.base_addr = binary_info.base_addr
            for provider in self.label_providers:
                provider.update(pdb_info)

    def resolveApi(self, to_address, api_address):
        for provider in self.label_providers:
            if not provider.isApiProvider():
                continue
            dll, api = provider.getApi(to_address, api_address)
            if dll or api:
                return (dll, api)

        return ("", "")

    def resolveSymbol(self, address):
        for provider in self.label_providers:
            if not provider.isSymbolProvider():
                continue
            result = provider.getSymbol(address)
            if result:
                return result
        return ""

    def getSymbolCandidates(self):
        symbol_offsets = set([])
        for provider in self.label_providers:
            if not provider.isSymbolProvider():
                continue
            function_symbols = provider.getFunctionSymbols()
            symbol_offsets.update(list(function_symbols.keys()))
        return list(symbol_offsets)

    def getReferencedAddr(self, op_str):
        referenced_addr = re.search(r"0x[a-fA-F0-9]+", op_str)
        if referenced_addr:
            return int(referenced_addr.group(), 16)
        return 0

    def resolveIndirectSwitch(self, addr_switch_array, size):
        indirect_switch_bytes = []
        current_offset = addr_switch_array + size * 4
        if self.disassembly.isAddrWithinMemoryImage(current_offset):
            LOGGER.debug("0x%08x analyzing potentially indirect switch table (size: 0x%08x).", current_offset, size)
            current_byte = self.disassembly.getByte(current_offset)
            if isinstance(current_byte, str):
                current_byte = ord(current_byte)
            while current_byte < size and not current_offset in self.fc_manager.getFunctionStartCandidates():
                indirect_switch_bytes.append(current_offset)
                current_offset += 1
                current_byte = self.disassembly.getByte(current_offset)
                if isinstance(current_byte, str):
                    current_byte = ord(current_byte)
            LOGGER.debug("0x%08x found %d bytes.", current_offset, len(indirect_switch_bytes))
        return indirect_switch_bytes

    def _analyzeCallInstruction(self, i, state):
        i_address, i_size, i_mnemonic, i_op_str = i
        state.setLeaf(False)
        # case = "FALLTHROUGH"
        call_destination = self.getReferencedAddr(i_op_str)
        if ":" in i_op_str:
            # case = "LONG-CALL"
            pass
        if i_op_str.startswith("dword ptr ["):
            # reg+offset is currently ignored as it is a minority of calls
            # case = "DWORD-PTR-REG"
            if i_op_str.startswith("dword ptr [0x"):
                # case = "DWORD-PTR"
                dereferenced = self.disassembly.dereferenceDword(call_destination)
                if dereferenced is not None:
                    state.addCodeRef(i_address, dereferenced)
                    self._handleCallTarget(state, i_address, dereferenced)
                    self._handleApiTarget(i_address, call_destination, dereferenced)
        elif i_op_str.startswith("qword ptr [rip"):
            rip = i_address + i_size
            call_destination = rip + self.getReferencedAddr(i_op_str)
            dereferenced = self.disassembly.dereferenceQword(call_destination)
            state.addCodeRef(i_address, call_destination)
            if dereferenced is not None:
                self._handleApiTarget(i_address, call_destination, dereferenced)
        elif i_op_str.startswith("0x"):
            # case = "DIRECT"
            self._handleCallTarget(state, i_address, call_destination)
            self._handleApiTarget(i_address, call_destination, call_destination)
        elif i_op_str.lower() in REGS_32BIT or i_op_str.lower() in REGS_64BIT:
            # case = "REG"
            # this is resolved by backtracking at the end of function analysis.
            state.call_register_ins.append(i_address)

    def _handleCallTarget(self, state, from_addr, to_addr):
        if to_addr and self.disassembly.isAddrWithinMemoryImage(to_addr):
            state.addCodeRef(from_addr, to_addr)
        if state.start_addr == to_addr:
            state.setRecursion(True)

    def _handleApiTarget(self, from_addr, to_addr, dereferenced):
        if to_addr:
            # identify API calls on the fly
            dll, api = self.resolveApi(to_addr, dereferenced)
            if dll or api:
                self._updateApiInformation(from_addr, dereferenced, dll, api)
                return (dll, api)
            elif not self.disassembly.isAddrWithinMemoryImage(to_addr):
                LOGGER.debug("potentially uncovered DLL address: 0x%08x", to_addr)

    def _updateApiInformation(self, from_addr, to_addr, dll, api):
        api_entry = {"referencing_addr": [], "dll_name": dll, "api_name": api}
        if to_addr in self.disassembly.apis:
            api_entry = self.disassembly.apis[to_addr]
        if from_addr not in api_entry["referencing_addr"]:
            api_entry["referencing_addr"].append(from_addr)
        self.disassembly.apis[to_addr] = api_entry

    def _analyzeCondJmpInstruction(self, i, state):
        i_address, i_size, i_mnemonic, i_op_str = i
        state.addBlockToQueue(i_address + i_size)
        jump_destination = self.getReferencedAddr(i_op_str)
        # case = "FALLTHROUGH"
        self.tailcall_analyzer.addJump(i_address, jump_destination)
        if jump_destination:
            if jump_destination in self.disassembly.functions:
                # case = "TAILCALL!"
                state.setSanelyEnding(True)
            elif jump_destination in self.fc_manager.getFunctionStartCandidates():
                # it's tough to decide whether this should be disassembled here or not. topic of "code-sharing functions".
                # case = "TAILCALL?"
                pass
            else:
                # case = "OFFSET-QUEUE"
                state.addBlockToQueue(int(i_op_str, 16))
            state.addCodeRef(i_address, int(i_op_str, 16), by_jump=True)
        state.setBlockEndingInstruction(True)

    def _analyzeLoopInstruction(self, i, state):
        i_address, i_size, i_mnemonic, i_op_str = i
        jump_destination = self.getReferencedAddr(i_op_str)
        if jump_destination:
            state.addCodeRef(i_address, int(i_op_str, 16), by_jump=True)
        # loops have two exits and should thus be handled as block ending instruction
        state.addBlockToQueue(i_address + i_size)
        state.setBlockEndingInstruction(True)

    def _analyzeJmpInstruction(self, i, state):
        i_address, i_size, i_mnemonic, i_op_str = i
        # case = "FALLTHROUGH"
        if ":" in i_op_str:
            # case = "LONG-JMP"
            pass
        elif i_op_str.startswith("dword ptr [0x"):
            # case = "DWORD-PTR"
            # Handles mostly jmp-to-api, stubs or tailcalls, all should be handled sanely this way.
            jump_destination = self.getReferencedAddr(i_op_str)
            dereferenced = self.disassembly.dereferenceDword(jump_destination)
            state.addCodeRef(i_address, jump_destination, by_jump=True)
            self.tailcall_analyzer.addJump(i_address, jump_destination)
            if dereferenced is not None:
                self._handleApiTarget(i_address, jump_destination, dereferenced)
        elif i_op_str.startswith("qword ptr [rip"):
            # case = "QWORD-PTR, RIP-relative"
            # Handles mostly jmp-to-api, stubs or tailcalls, all should be handled sanely this way.
            rip = i_address + i_size
            jump_destination = rip + self.getReferencedAddr(i_op_str)
            dereferenced = self.disassembly.dereferenceQword(jump_destination)
            state.addCodeRef(i_address, jump_destination, by_jump=True)
            self.tailcall_analyzer.addJump(i_address, jump_destination)
            if dereferenced is not None:
                self._handleApiTarget(i_address, jump_destination, dereferenced)
        elif i_op_str.startswith("0x"):
            jump_destination = self.getReferencedAddr(i_op_str)
            self.tailcall_analyzer.addJump(i_address, jump_destination)
            if jump_destination in self.disassembly.functions:
                # case = "TAILCALL!"
                state.setSanelyEnding(True)
            elif jump_destination in self.fc_manager.getFunctionStartCandidates():
                # case = "TAILCALL?"
                pass
            else:
                if state.isFirstInstruction():
                    # case = "STUB-TAILCALL!"
                    pass
                else:
                    # case = "OFFSET-QUEUE"
                    state.addBlockToQueue(int(i_op_str, 16))
            state.addCodeRef(i_address, int(i_op_str, 16), by_jump=True)
        else:
            jumptable_targets = self.jumptable_analyzer.getJumpTargets(i, state)
            for target in jumptable_targets:
                if self.disassembly.isAddrWithinMemoryImage(target):
                    state.addBlockToQueue(target)
                    state.addCodeRef(i_address, target, by_jump=True)
        state.setNextInstructionReachable(False)
        state.setBlockEndingInstruction(True)

    def _analyzeEndInstruction(self, state):
        state.setSanelyEnding(True)
        state.setNextInstructionReachable(False)
        state.setBlockEndingInstruction(True)

    def _getDisasmWindowBuffer(self, addr):
        relative_start = addr - self.disassembly.binary_info.base_addr
        relative_end = relative_start + 15
        return self.disassembly.binary_info.binary[relative_start:relative_end]

    def analyzeFunction(self, start_addr, as_gap=False):
        LOGGER.debug("analyzeFunction() starting analysis of candidate @0x%08x", start_addr)
        self.tailcall_analyzer.initFunction()
        i = None
        state = FunctionAnalysisState(start_addr, self.disassembly)
        if state.isProcessedFunction():
            self.fc_manager.updateAnalysisAborted(start_addr, "collision with existing code of function 0x{:08x}".format(self.disassembly.ins2fn[start_addr]))
            return []
        while state.hasUnprocessedBlocks():
            LOGGER.debug("  current block queue: %s", ", ".join(["0x%x" % addr for addr in state.block_queue]))
            state.chooseNextBlock()
            LOGGER.debug("  analyzeFunction() now processing block @0x%08x", state.block_start)
            # in capstone, disassembly is more expensive than calling the function, so we use maximum x86/64 instruction size (14 bytes) as lookeahead.
            # disasm_lite() also provides up to 30% faster disassembly than disasm(), so we work with tuples instead of objects
            cache = [i for i in self.capstone.disasm_lite(self._getDisasmWindowBuffer(state.block_start), state.block_start)]
            cache_pos = 0
            previous_address = None
            previous_mnemonic = None
            previous_op_str = None
            while True:
                for i in cache:
                    i_address, i_size, i_mnemonic, i_op_str = i
                    i_op_str = i_op_str.strip()
                    i_relative_address = i_address - self.disassembly.binary_info.base_addr
                    i_bytes = self.disassembly.binary_info.binary[i_relative_address:i_relative_address + i_size]
                    LOGGER.debug("  analyzeFunction() now processing instruction @0x%08x: %s", i_address, i_mnemonic + " " + i_op_str)
                    cache_pos += i_size
                    state.setNextInstructionReachable(True)
                    # count appearences of "suspicious" byte patterns (like 00 00) that indicate non-function code
                    if i_bytes == DOUBLE_ZERO:
                        state.suspicious_ins_count += 1
                        LOGGER.debug("    analyzeFunction() found suspicious function @0x%08x", i_address)
                        if state.suspicious_ins_count > 1:
                            self.fc_manager.updateAnalysisAborted(start_addr, "too many suspicious instructions @0x%08x" % i_address)
                            return state
                    if i_mnemonic in CALL_INS:
                        self._analyzeCallInstruction(i, state)
                    elif i_mnemonic in JMP_INS:
                        self._analyzeJmpInstruction(i, state)
                    elif i_mnemonic in LOOP_INS:
                        self._analyzeLoopInstruction(i, state)
                    elif i_mnemonic in CJMP_INS:
                        self._analyzeCondJmpInstruction(i, state)
                    elif i_mnemonic.startswith("j"):
                        LOGGER.error("unsupported jump @0x%08x (0x%08x): %s %s", i_address, start_addr, i_mnemonic, i_op_str)
                        # we do not analyze any potential exception handler (tricks), so treat breakpoints as exit condition
                    elif i_mnemonic in RET_INS:
                        self._analyzeEndInstruction(state)
                        LOGGER.debug("  analyzeFunction() found ending instruction @0x%08x", i_address)
                        if previous_address and previous_mnemonic == "push":
                            push_ret_destination = self.getReferencedAddr(previous_op_str)
                            if self.disassembly.isAddrWithinMemoryImage(push_ret_destination):
                                LOGGER.debug("  analyzeFunction() found push-return jump obfuscation: @0x%08x", i_address)
                                state.addBlockToQueue(push_ret_destination)
                                state.addCodeRef(i_address, push_ret_destination, by_jump=True)
                    elif i_mnemonic in ["int3", "hlt"]:
                        self._analyzeEndInstruction(state)
                        LOGGER.debug("  analyzeFunction() found ending instruction @0x%08x", i_address)
                    elif previous_address and i_address != start_addr and previous_mnemonic == "call":
                        instruction_sequence = [ins for ins in self.capstone.disasm(self._getDisasmWindowBuffer(i_address), i_address)]
                        if (not self.disassembly.language['_guess'] == "go" and self.fc_manager.isAlignmentSequence(instruction_sequence)) or self.fc_manager.isFunctionCandidate(i_address):
                            # LLVM and GCC sometimes tends to produce lots of tailcalls that basically mess with function end detection, we cut whenever we find effective nops after calls
                            # however, Go tends to insert alignment NOPs after calls, too, but in this case, they are no tailcall indicator
                            LOGGER.debug("    current function: 0x%x ---> ran into alignment sequence after call -> 0x%08x, cutting block here.", start_addr, i_address)
                            state.setBlockEndingInstruction(True)
                            state.endBlock()
                            state.setSanelyEnding(True)
                            if self.fc_manager.isAlignmentSequence(instruction_sequence):
                                next_aligned_address = previous_address + (16 - previous_address % 16)
                                LOGGER.debug("  Adding: 0x%x as candidate.", next_aligned_address)
                                self.fc_manager.addCandidate(next_aligned_address, is_gap=True)
                            break
                    previous_address = i_address
                    previous_mnemonic = i_mnemonic
                    previous_op_str = i_op_str
                    if not i_address in self.disassembly.code_map and not i_address in self.disassembly.data_map and not state.isProcessed(i_address):
                        LOGGER.debug("  analyzeFunction() booked instruction @0x%08x: %s for processed state", i_address, i_mnemonic + " " + i_op_str)
                        state.addInstruction(i_address, i_size, i_mnemonic, i_op_str, i_bytes)
                    elif i_address in self.disassembly.code_map:
                        LOGGER.debug("  analyzeFunction() was already present?! instruction @0x%08x: %s (function: 0x%08x)", i_address, i_mnemonic + " " + i_op_str, self.disassembly.ins2fn[i_address])
                        state.setBlockEndingInstruction(True)
                        state.setCollision(True)
                    else:
                        LOGGER.debug("  analyzeFunction() was already present in local function.")
                        state.setBlockEndingInstruction(True)
                    if state.isBlockEndingInstruction():
                        state.endBlock()
                        break
                else:
                    #if the inner loop did not break, we need to refill the cache in order to finish the block-analysis
                    cache = [i for i in self.capstone.disasm_lite(self._getDisasmWindowBuffer(state.block_start + cache_pos), state.block_start + cache_pos)]
                    if not cache:
                        break
                    continue
                #if the inner loop did break, the cache didn't run empty and thus block-analysis is finished
                break
            if not state.isBlockEndingInstruction():
                if i is not None:
                    LOGGER.debug("No block submitted, last instruction: 0x%08x -> 0x%08x %s || %s",
                                 start_addr,
                                 i_address,
                                 i_mnemonic + " " + i_op_str,
                                 self.fc_manager.getFunctionCandidate(start_addr))
                else:
                    LOGGER.debug("No block submitted with no ins, last instruction: 0x%08x || %s",
                                 start_addr,
                                 self.fc_manager.getFunctionCandidate(start_addr))
        state.label = self.resolveSymbol(state.start_addr)
        analysis_result = state.finalizeAnalysis(as_gap)
        if analysis_result and self.config.RESOLVE_REGISTER_CALLS:
            self.indcall_analyzer.resolveRegisterCalls(state)
            self.tailcall_analyzer.finalizeFunction(state)
        self.fc_manager.updateAnalysisFinished(start_addr)
        self.fc_manager.updateCandidates(state)
        return state

    def analyzeBuffer(self, binary_info, cbAnalysisTimeout=None):
        LOGGER.debug("Analyzing buffer with %d bytes @0x%08x", binary_info.binary_size, binary_info.base_addr)
        self._updateLabelProviders(binary_info)
        self.disassembly = DisassemblyResult()
        self.disassembly.smda_version = self.config.VERSION
        self.disassembly.setBinaryInfo(binary_info)
        self.disassembly.binary_info.architecture = "intel"
        self.disassembly.analysis_start_ts = datetime.datetime.utcnow()
        if self.disassembly.binary_info.bitness not in [32, 64]:
            bitness_analyzer = BitnessAnalyzer()
            self.disassembly.binary_info.bitness = bitness_analyzer.determineBitnessFromDisassembly(self.disassembly)
            LOGGER.debug("Automatically Recognized Bitness as: %d", self.disassembly.binary_info.bitness)
        else:
            LOGGER.debug("Using defined Bitness as: %d", self.disassembly.binary_info.bitness)
        if self._forced_bitness:
            self.disassembly.binary_info.bitness = self._forced_bitness
            LOGGER.debug("Forced Bitness override to: %d", self.disassembly.binary_info.bitness)

        self.tailcall_analyzer = TailcallAnalyzer()
        self.indcall_analyzer = IndirectCallAnalyzer(self)
        self.jumptable_analyzer = JumpTableAnalyzer(self)

        self.fc_manager = FunctionCandidateManager(self.config)
        if self.config.USE_SYMBOLS_AS_CANDIDATES:
            self.fc_manager.symbol_addresses = self.getSymbolCandidates()
        self.fc_manager.init(self.disassembly)
        self._initCapstone()
        self._initTfIdf()
        # first pass, analyze locations identifiable by heuristics (e.g. call-reference, common prologue)
        for candidate in self.fc_manager.getNextFunctionStartCandidate():
            if cbAnalysisTimeout and cbAnalysisTimeout():
                break
            state = self.analyzeFunction(candidate.addr)
        LOGGER.debug("Finished heuristical analysis, functions: %d", len(self.disassembly.functions))
        # second pass, analyze remaining gaps for additional candidates in an iterative way
        gap_candidate = self.fc_manager.nextGapCandidate()
        while gap_candidate is not None:
            if cbAnalysisTimeout and cbAnalysisTimeout():
                break
            LOGGER.debug("based on gap, performing function analysis of 0x%08x", gap_candidate)
            state = self.analyzeFunction(gap_candidate, as_gap=True)
            function_blocks = state.getBlocks()
            if function_blocks:
                LOGGER.debug("+ got some blocks here -> 0x%08x", gap_candidate)
            if gap_candidate in self.disassembly.functions:
                fn_min = self.disassembly.function_borders[gap_candidate][0]
                fn_max = self.disassembly.function_borders[gap_candidate][1]
                LOGGER.debug("+++ YAY, is now a function! -> 0x%08x - 0x%08x", fn_min, fn_max)
                # start looking directly after our new function
            else:
                self.fc_manager.updateAnalysisAborted(gap_candidate, "Gap candidate did not fulfil function criteria.")
            next_gap = self.fc_manager.getNextGap(dont_skip=True)
            gap_candidate = self.fc_manager.nextGapCandidate(next_gap)
        LOGGER.debug("Finished gap analysis, functions: %d", len(self.disassembly.functions))
        # third pass, fix potential tailcall functions that were identified during analysis
        if self.config.RESOLVE_TAILCALLS:
            tailcalled_functions = self.tailcall_analyzer.resolveTailcalls(self)
            for addr in tailcalled_functions:
                self.fc_manager.addTailcallCandidate(addr)
            LOGGER.debug("Finished tailcall analysis, functions.")
        self.disassembly.failed_analysis_addr = self.fc_manager.getAbortedCandidates()
        # package up and finish
        for addr, candidate in self.fc_manager.candidates.items():
            if addr in self.disassembly.functions:
                function_blocks = self.disassembly.getBlocksAsDict(addr)
                function_tfidf = self._tfidf.getTfIdfFromBlocks(function_blocks)
                candidate.setTfIdf(function_tfidf)
                candidate.getConfidence()
            self.disassembly.candidates[addr] = candidate
        self.disassembly.analysis_end_ts = datetime.datetime.utcnow()
        if cbAnalysisTimeout():
            self.disassembly.analysis_timeout = True
        return self.disassembly



#!/usr/bin/env python3
import hashlib
import struct

from smda.common.Tarjan import Tarjan
from smda.common.CodeXref import CodeXref
from smda.common.SmdaBasicBlock import SmdaBasicBlock
from smda.common.DominatorTree import build_dominator_tree, get_nesting_depth
from smda.intel.IntelInstructionEscaper import IntelInstructionEscaper
from .SmdaInstruction import SmdaInstruction


class SmdaFunction(object):

    smda_report = None
    offset = None
    blocks = None
    apirefs = None
    blockrefs = None
    inrefs = None
    outrefs = None
    code_inrefs = None
    code_outrefs = None
    is_exported = None
    # metadata
    binweight = 0
    characteristics = ""
    confidence = 0.0
    function_name = ""
    pic_hash = None
    strongly_connected_components = None
    tfidf = None

    def __init__(self, disassembly=None, function_offset=None, config=None, smda_report=None):
        self.smda_report = smda_report
        if disassembly is not None and function_offset is not None:
            self._escaper = IntelInstructionEscaper if disassembly.binary_info.architecture in ["intel"] else None
            self.offset = function_offset
            self._parseBlocks(disassembly.getBlocksAsDict(function_offset))
            self.apirefs = disassembly.getApiRefs(function_offset)
            self.blockrefs = disassembly.getBlockRefs(function_offset)
            self.inrefs = disassembly.getInRefs(function_offset)
            self.outrefs = disassembly.getOutRefs(function_offset)
            self.is_exported = self.offset in disassembly.exported_functions
            # metadata
            self.function_name = disassembly.function_symbols.get(function_offset, "")
            self.characteristics = disassembly.candidates[function_offset].getCharacteristics() if function_offset in disassembly.candidates else None
            self.confidence = disassembly.candidates[function_offset].getConfidence() if function_offset in disassembly.candidates else None
            self.tfidf = disassembly.candidates[function_offset].getTfIdf() if function_offset in disassembly.candidates else None
            self.pic_hash = self._calculatePicHash(disassembly.binary_info)
            if config and config.CALCULATE_SCC:
                self.strongly_connected_components = self._calculateSccs()
            if config and config.CALCULATE_NESTING:
                self.nesting_depth = self._calculateNestingDepth()

    @property
    def num_edges(self):
        return sum([len(value) for key, value in self.blockrefs.items()])

    @property
    def num_inrefs(self):
        return len(self.inrefs)

    @property
    def num_outrefs(self):
        return sum([len(dsts) for src, dsts in self.outrefs.items()])

    @property
    def num_blocks(self):
        return len(self.blocks)

    @property
    def num_instructions(self):
        return sum([1 for ins in self.getInstructions()])

    @property
    def num_calls(self):
        return sum([1 for ins in self.getInstructions() if ins.mnemonic == "call"])

    @property
    def num_returns(self):
        return sum([1 for ins in self.getInstructions() if ins.mnemonic in ["ret", "retn"]])

    def isApiThunk(self):
        if self.num_instructions != 1:
            return False
        first_ins = self.blocks[self.offset][0]
        if first_ins.mnemonic not in ["jmp", "call"]:
            return False
        if len(self.apirefs) == 0:
            return False
        return True

    def isExported(self):
        return self.is_exported

    def getBlocks(self):
        for _, block in sorted(self.blocks.items()):
            yield SmdaBasicBlock(block, smda_function=self)

    def getPicHashAsLong(self):
        return self.pic_hash

    def getPicHashAsHex(self):
        return struct.pack("l", self.pic_hash).hex()

    def getInstructions(self):
        for block in self.getBlocks():
            for ins in block.getInstructions():
                yield ins

    def getInstructionsForBlock(self, offset):
        if offset is None:
            offset = self.offset
        if offset not in self.blocks:
            return []
        return self.blocks[offset]

    def getCodeInrefs(self):
        self.smda_report.initCodeXrefs()
        # potentially lazy initialize CodeXrefs externally via SmdaReport
        for inref in self.code_inrefs:
            yield inref

    def getCodeOutrefs(self):
        self.smda_report.initCodeXrefs()
        # potentially lazy initialize CodeXrefs externally via SmdaReport
        for outref in self.code_outrefs:
            yield outref

    def _calculateSccs(self):
        tarjan = Tarjan(self.blockrefs)
        tarjan.calculateScc()
        return tarjan.getResult()

    def _calculateNestingDepth(self):
        nesting_depth = 0
        try:
            if self.blockrefs:
                tree = build_dominator_tree(self.blockrefs, self.offset)
                if tree:
                    nesting_depth = get_nesting_depth(self.blockrefs, tree, self.offset)
        except:
            pass
        return nesting_depth

    def _calculatePicHash(self, binary_info):
        escaped_binary_seqs = []
        for _, block in sorted(self.blocks.items()):
            for instruction in block:
                escaped_binary_seqs.append(instruction.getEscapedBinary(self._escaper, lower_addr=binary_info.base_addr, upper_addr=binary_info.base_addr + binary_info.binary_size))
        as_bytes = bytes([ord(c) for c in "".join(escaped_binary_seqs)])
        return struct.unpack("Q", hashlib.sha256(as_bytes).digest()[:8])[0]

    def _parseBlocks(self, block_dict):
        self.blocks = {}
        for offset, block in block_dict.items():
            instructions = [SmdaInstruction(ins, smda_function=self) for ins in block]
            self.blocks[int(offset)] = instructions
            self.binweight += sum([len(ins.bytes) / 2 for ins in instructions])

    @classmethod
    def fromDict(cls, function_dict, binary_info=None, version=None, smda_report=None):
        smda_function = cls(None)
        smda_function.smda_report = smda_report
        smda_function.offset = function_dict["offset"]
        smda_function.blocks = {}
        for addr, block in function_dict["blocks"].items():
            smda_function.blocks[int(addr)] = [SmdaInstruction.fromDict(ins) for ins in block]
        smda_function.apirefs = {int(k): v for k, v in function_dict["apirefs"].items()}
        smda_function.blockrefs = {int(k): v for k, v in function_dict["blockrefs"].items()}
        smda_function.inrefs = function_dict["inrefs"]
        smda_function.outrefs = {int(k): v for k, v in function_dict["outrefs"].items()}
        # provide some legacy support by assuming functions are not exported for SMDA reports < 1.7.0
        smda_function.is_exported = function_dict["is_exported"] if "is_exported" in function_dict else False
        smda_function.binweight = function_dict["metadata"]["binweight"]
        smda_function.characteristics = function_dict["metadata"]["characteristics"]
        smda_function.confidence = function_dict["metadata"]["confidence"]
        smda_function.function_name = function_dict["metadata"]["function_name"]
        smda_function.pic_hash = function_dict["metadata"]["pic_hash"]
        smda_function.strongly_connected_components = function_dict["metadata"]["strongly_connected_components"]
        smda_function.tfidf = function_dict["metadata"]["tfidf"]
        if binary_info.architecture:
            smda_function._escaper = IntelInstructionEscaper if binary_info.architecture in ["intel"] else None
        # modernize older reports on import
        if version and version.startswith("1.2"):
            smda_function.nesting_depth = smda_function._calculateNestingDepth()
            smda_function.pic_hash = smda_function._calculatePicHash(binary_info)
        else:
            smda_function.nesting_depth = function_dict["metadata"]["nesting_depth"]
        return smda_function

    def toDict(self):
        blocks_as_dict = {}
        for addr, block in self.blocks.items():
            blocks_as_dict[addr] = [ins.toDict() for ins in block]
        return  {
            "offset": self.offset,
            "blocks": blocks_as_dict,
            "apirefs": self.apirefs,
            "blockrefs": self.blockrefs,
            "inrefs": self.inrefs,
            "outrefs": self.outrefs,
            "is_exported": self.is_exported,
            "metadata": {
                "binweight": self.binweight,
                "characteristics": self.characteristics,
                "confidence": self.confidence,
                "function_name": self.function_name,
                "pic_hash": self.pic_hash,
                "nesting_depth": self.nesting_depth,
                "strongly_connected_components": self.strongly_connected_components,
                "tfidf": self.tfidf
            }
        }

    def __int__(self):
        return self.offset

    def __str__(self):
        return "0x{:08x}: (->{:>4d}, {:>4d}->) {:>3d} blocks, {:>4d} instructions.".format(self.offset, self.num_inrefs, self.num_outrefs, self.num_blocks, self.num_instructions)



from smda.utility.FileLoader import FileLoader


class MemoryFileLoader(FileLoader):

    def __init__(self, buffer, load_file=True, map_file=False):
        super().__init__("", load_file=False, map_file=map_file)
        self._loadFile(buffer=buffer)



import re
import struct
import logging

from capstone import Cs, CS_ARCH_X86, CS_MODE_32, CS_MODE_64

from smda.utility.PriorityQueue import PriorityQueue
from .definitions import DEFAULT_PROLOGUES, GAP_SEQUENCES
from .LanguageAnalyzer import LanguageAnalyzer
from .FunctionCandidate import FunctionCandidate

LOGGER = logging.getLogger(__name__)


class FunctionCandidateManager(object):

    def __init__(self, config):
        self.config = config
        self.lang_analyzer = None
        self.disassembly = None
        self.bitness = None
        self._code_areas = []
        self.candidates = {}
        self.candidate_queue = []
        self.cached_candidates = None
        self._candidate_offsets = []
        self.candidate_index = 0
        self._all_call_refs = {}
        self.symbol_addresses = []
        self.identified_alignment = 0
        self.go_objects = None
        # gap filling
        self.function_gaps = None
        self.max_function_addr = 0
        self.gap_pointer = None
        self.previously_analyzed_gap = 0
        self.capstone = None

    def init(self, disassembly):
        if disassembly.binary_info.code_areas:
            self._code_areas = disassembly.binary_info.code_areas
        self.disassembly = disassembly
        self.lang_analyzer = LanguageAnalyzer(disassembly)
        self.disassembly.language = self.lang_analyzer.identify()
        self.bitness = disassembly.binary_info.bitness
        self.capstone = Cs(CS_ARCH_X86, CS_MODE_32)
        if self.bitness == 64:
            self.capstone = Cs(CS_ARCH_X86, CS_MODE_64)
        self.locateCandidates()
        self.disassembly.identified_alignment = self.identified_alignment
        self._buildQueue()

    def _passesCodeFilter(self, addr):
        if addr is None:
            return False
        if self._code_areas:
            for area in self._code_areas:
                if area[0] <= addr < area[1]:
                    return True
            return False
        return True

    def getBitMask(self):
        if self.bitness == 64:
            return 0xFFFFFFFFFFFFFFFF
        return 0xFFFFFFFF

    def setInitialCandidate(self, addr):
        if addr in self.candidates:
            self.candidates[addr].setInitialCandidate(True)

    def isFunctionCandidate(self, addr):
        return addr in self.candidates

    def getFunctionCandidate(self, addr):
        if addr in self.candidates:
            return self.candidates[addr]
        return None

    def getAbortedCandidates(self):
        aborted = []
        for addr, candidate in self.candidates.items():
            if candidate.analysis_aborted:
                aborted.append(addr)
        return aborted

    def updateAnalysisAborted(self, addr, reason):
        LOGGER.debug("function analysis of 0x%08x aborted: %s", addr, reason)
        if addr in self.candidates:
            self.candidates[addr].setAnalysisAborted(reason)

    def updateAnalysisFinished(self, addr):
        LOGGER.debug("function analysis of 0x%08x successfully completed.", addr)
        if addr in self.candidates:
            self.candidates[addr].setAnalysisCompleted()

    def updateCandidates(self, state):
        if self.config.HIGH_ACCURACY:
            conflicts = state.identifyCallConflicts(self._all_call_refs)
            if conflicts:
                for candidate_addr, conflict in conflicts.items():
                    self.candidates[candidate_addr].removeCallRefs(conflict)
                self.candidate_queue.update()

    def addCandidate(self, addr, is_gap=False, reference_source=None):
        if not self._passesCodeFilter(addr):
            return False
        self.ensureCandidate(addr)
        self.candidates[addr].setIsGapCandidate(is_gap)
        if reference_source:
            self.candidates[addr].addCallRef(reference_source)
        self.candidate_queue.add(self.candidates[addr])
        self.candidate_queue.update()

    def getNextFunctionStartCandidate(self):
        for candidate in self.candidate_queue:
            if not (candidate.isFinished() or candidate.getScore() == 0):
                if self.identified_alignment and candidate.alignment < self.identified_alignment:
                    continue
                yield candidate

    def _logCandidateStats(self):
        LOGGER.debug("Candidate Statistics:")
        try:
            maxc = max([c.getScore() for c in self.candidates.values()])
            minc = min([c.getScore() for c in self.candidates.values()])
            candidates_2 = len([c.getScore() for c in self.candidates.values() if c.getScore() == 2])
            candidates_1 = len([c.getScore() for c in self.candidates.values() if c.getScore() == 1])
            candidates_0 = len([c.getScore() for c in self.candidates.values() if c.getScore() == 0])
            LOGGER.debug("  Max: %f, Min: %f", maxc, minc)
            LOGGER.debug("  2: %d, 1: %d, 0: %d", candidates_2, candidates_1, candidates_0)
        except:
            LOGGER.debug("  No candidates found.")

    def getFunctionStartCandidates(self):
        return self._candidate_offsets

    def updateFunctionGaps(self):
        gaps = []
        prev_ins = 0
        min_code = min(self.disassembly.code_map) if self.disassembly.code_map else self.getBitMask()
        max_code = max(self.disassembly.code_map) if self.disassembly.code_map else 0
        for code_area in self._code_areas:
            if code_area[0] < min_code < code_area[1] and min_code != code_area[0]:
                gaps.append([code_area[0], min_code, min_code - code_area[0]])
            if code_area[0] < max_code < code_area[1] and max_code != code_area[1]:
                gaps.append([max_code, code_area[1], code_area[1] - max_code])
        for ins in sorted(self.disassembly.code_map.keys()):
            if prev_ins != 0:
                if ins - prev_ins > 1:
                    gaps.append([prev_ins + 1, ins, ins - prev_ins])
            prev_ins = ins
        self.function_gaps = sorted(gaps)

    def initGapSearch(self):
        if self.gap_pointer is None:
            LOGGER.debug("initGapSearch()")
            self.gap_pointer = self.getBitMask()
            self.updateFunctionGaps()
            if self.function_gaps:
                self.gap_pointer = self.function_gaps[0][0]
        LOGGER.debug("initGapSearch() gaps are:")
        for gap in self.function_gaps:
            LOGGER.debug("initGapSearch() 0x%08x - 0x%08x == %d", gap[0], gap[1], gap[2])
        return

    def getNextGap(self, dont_skip=False):
        next_gap = self.getBitMask()
        for gap in self.function_gaps:
            if gap[0] > self.gap_pointer:
                next_gap = gap[0]
                break
        LOGGER.debug("getNextGap(%s) for 0x%08x based on gap_map: 0x%08x", dont_skip, self.gap_pointer, next_gap)
        # we potentially just disassembled a function and want to continue directly behind it in case we would otherwise miss more
        if dont_skip:
            if self.gap_pointer in self.disassembly.code_map:
                function = self.disassembly.ins2fn[self.gap_pointer]
                next_gap = min(next_gap, self.disassembly.function_borders[function][1])
                LOGGER.debug("getNextGap(%s) without skip => after checking versus code map: 0x%08x", dont_skip, next_gap)
        LOGGER.debug("getNextGap(%s) final gap_ptr: 0x%08x", dont_skip, next_gap)
        return next_gap

    def isEffectiveNop(self, byte_sequence):
        if byte_sequence in GAP_SEQUENCES[len(byte_sequence)]:
            return True
        return False

    def isAlignmentSequence(self, instruction_sequence):
        is_alignment_sequence = False
        if len(instruction_sequence) > 0:
            current_offset = instruction_sequence[0].address
            for instruction in instruction_sequence:
                if instruction.bytes in GAP_SEQUENCES[len(instruction.bytes)]:
                    current_offset += len(instruction.bytes)
                    if current_offset % 16 == 0:
                        is_alignment_sequence = True
                        break
                else:
                    break
        return is_alignment_sequence

    def nextGapCandidate(self, start_gap_pointer=None):
        if self.gap_pointer is None:
            self.initGapSearch()
        if start_gap_pointer:
            self.gap_pointer = start_gap_pointer
        LOGGER.debug("nextGapCandidate() finding new gap candidate, current gap_ptr: 0x%08x", self.gap_pointer)
        while True:
            if self.disassembly.binary_info.base_addr + self.disassembly.binary_info.binary_size < self.gap_pointer:
                LOGGER.debug("nextGapCandidate() gap_ptr: 0x%08x - finishing", self.gap_pointer)
                return None
            gap_offset = self.gap_pointer - self.disassembly.binary_info.base_addr
            if gap_offset >= self.disassembly.binary_info.binary_size:
                return None
            # compatibility with python2/3...
            try:
                byte = self.disassembly.getRawByte(gap_offset)
            except:
                pass
                LOGGER.warn("could not fetch raw byte for gap pointer.")
                # print("0x%08x" % self.disassembly.binary_info.base_addr, "0x%08x" % self.disassembly.binary_info.binary_size, "0x%08x" % self.gap_pointer, "0x%08x" % gap_offset)
            # try to find padding symbols and skip them
            if isinstance(byte, int):
                byte = struct.pack("B", byte)
            if byte in GAP_SEQUENCES[1]:
                LOGGER.debug("nextGapCandidate() found 0xCC / 0x00 - gap_ptr += 1: 0x%08x", self.gap_pointer)
                self.gap_pointer += 1
                continue
            # try to find instructions that directly encode as NOP and skip them
            ins_buf = [i for i in self.capstone.disasm_lite(self.disassembly.getRawBytes(gap_offset, 15), gap_offset)]
            if ins_buf:
                i_address, i_size, i_mnemonic, i_op_str = ins_buf[0]
                if  i_mnemonic == "nop":
                    nop_instruction = i_mnemonic + " " + i_op_str
                    nop_length = i_size
                    LOGGER.debug("nextGapCandidate() found nop instruction (%s) - gap_ptr += %d: 0x%08x", nop_instruction, nop_length, self.gap_pointer)
                    self.gap_pointer += nop_length
                    continue
            # try to find effective NOPs and skip them.
            found_multi_byte_nop = False
            for gap_length in range(max(GAP_SEQUENCES.keys()), 1, -1):
                if self.disassembly.getRawBytes(gap_offset, gap_length) in GAP_SEQUENCES[gap_length]:
                    LOGGER.debug("nextGapCandidate() found %d byte effective nop - gap_ptr += %d: 0x%08x", gap_length, gap_length, self.gap_pointer)
                    self.gap_pointer += gap_length
                    found_multi_byte_nop = True
                    break
            if found_multi_byte_nop:
                continue
            # we know this place from data already
            if self.gap_pointer in self.disassembly.data_map:
                LOGGER.debug("nextGapCandidate() gap_ptr is already inside data map: 0x%08x", self.gap_pointer)
                self.gap_pointer += 1
                continue
            if self.gap_pointer in self.disassembly.code_map:
                LOGGER.debug("nextGapCandidate() gap_ptr is already inside code map: 0x%08x", self.gap_pointer)
                self.gap_pointer = self.getNextGap()
                continue
            # we may have a candidate here
            LOGGER.debug("nextGapCandidate() using 0x%08x as candidate", self.gap_pointer)
            start_byte = self.disassembly.getRawByte(gap_offset)
            has_common_prologue = True  # start_byte in FunctionCandidate(self.gap_pointer, start_byte, self.bitness).common_gap_starts[self.bitness]
            if self.previously_analyzed_gap == self.gap_pointer:
                LOGGER.debug("--- HRM, nextGapCandidate() gap_ptr at: 0x%08x was previously analyzed", self.gap_pointer)
                self.gap_pointer = self.getNextGap(dont_skip=True)
            elif not has_common_prologue:
                LOGGER.debug("--- HRM, nextGapCandidate() gap_ptr at: 0x%08x has no common prologue (0x%08x)", self.gap_pointer, ord(start_byte))
                self.gap_pointer = self.getNextGap(dont_skip=True)
            else:
                self.previously_analyzed_gap = self.gap_pointer
                self.addGapCandidate(self.gap_pointer)
                return self.gap_pointer
        return None

    def checkFunctionOverlap(self):
        function_boundaries = []
        for function in self.disassembly.functions:
            min_addr = self.getBitMask()
            max_addr = 0
            for block in self.disassembly.functions[function]:
                min_addr = min(min_addr, min([instruction[0] for instruction in block]))
                max_addr = max(max_addr, max([instruction[0] + instruction[1] for instruction in block]))
            function_boundaries.append((min_addr, max_addr))
        current_entry = (0, 0)
        for entry in sorted(function_boundaries):
            if current_entry[1] > entry[0]:
                return True
            current_entry = entry
        return False

    def checkCodePadding(self):
        pattern_count = 0
        pattern_functions = []
        for pattern in re.finditer(r"((\xCC){2,}|(\x90){2,})", self.disassembly.binary_info.binary):
            pattern_count += 1
            pattern_functions.append(pattern.span()[1] + 1)

    def ensureCandidate(self, addr):
        """ create candidate if it does not exist yet, returns True if newly created, else False """
        if addr not in self.candidates:
            self.candidates[addr] = FunctionCandidate(self.disassembly.binary_info, addr)
            return True
        return False

    def addGapCandidate(self, addr):
        if not self._passesCodeFilter(addr):
            return False
        self.ensureCandidate(addr)
        self.candidates[addr].setIsGapCandidate(True)

    def addTailcallCandidate(self, addr):
        if not self._passesCodeFilter(addr):
            return False
        self.ensureCandidate(addr)
        self.candidates[addr].setIsTailcallCandidate(True)

    def addReferenceCandidate(self, addr, source_ref):
        if not self._passesCodeFilter(addr):
            return False
        if self.ensureCandidate(addr):
            self._all_call_refs[source_ref] = addr
        self.candidates[addr].addCallRef(source_ref)

    def addLanguageSpecCandidate(self, addr, lang_spec):
        if not self._passesCodeFilter(addr):
            return False
        self.ensureCandidate(addr)
        self.candidates[addr].setLanguageSpec(lang_spec)

    def addPrologueCandidate(self, addr):
        if not self._passesCodeFilter(addr):
            return False
        return self.ensureCandidate(addr)

    def addSymbolCandidate(self, addr):
        if not self._passesCodeFilter(addr):
            return False
        self.ensureCandidate(addr)
        self.candidates[addr].setIsSymbol(True)
        self.candidates[addr].setInitialCandidate(True)

    def addExceptionCandidate(self, addr):
        if not self._passesCodeFilter(addr):
            return False
        self.ensureCandidate(addr)
        self.candidates[addr].setIsExceptionHandler(True)
        self.candidates[addr].setInitialCandidate(True)

    def resolvePointerReference(self, offset):
        if self.bitness == 32:
            addr_block = self.disassembly.getRawBytes(offset + 2, 4)
            function_pointer = struct.unpack("I", addr_block)[0]
            return self.disassembly.dereferenceDword(function_pointer)
        if self.bitness == 64:
            addr_block = self.disassembly.getRawBytes(offset + 2, 4)
            function_pointer = struct.unpack("i", addr_block)[0]
            # we need to calculate RIP + offset + 7 (48 ff 25 ** ** ** **)
            if self.disassembly.getRawBytes(offset, 2) == "\xFF\x25":
                function_pointer += offset + 7
            elif self.disassembly.getRawBytes(offset, 2) == "\xFF\x15":
                function_pointer += offset + 6
            else:
                raise Exception("resolvePointerReference: should only be used on call/jmp * ptr")
            return self.disassembly.binary_info.base_addr + function_pointer
        raise Exception("resolvePointerReference: undefined bitness")

    def _identifyAlignment(self):
        identified_alignment = 0
        if self.config.USE_ALIGNMENT:
            num_candidates = sum([1 for addr, candidate in self.candidates.items() if len(candidate.call_ref_sources) > 1])
            num_aligned_16_candidates = sum([1 for addr, candidate in self.candidates.items() if len(candidate.call_ref_sources) > 1 and candidate.alignment == 16])
            num_aligned_4_candidates = sum([1 for addr, candidate in self.candidates.items() if len(candidate.call_ref_sources) > 1 and candidate.alignment >= 4])
            if num_candidates:
                alignment_16_ratio = 1.0 * num_aligned_16_candidates / num_candidates
                alignment_4_ratio = 1.0 * num_aligned_4_candidates / num_candidates
                if num_candidates > 20 and alignment_4_ratio > 0.95:
                    identified_alignment = 4
                if num_candidates > 20 and alignment_16_ratio > 0.95:
                    identified_alignment = 16
        return identified_alignment

    def locateCandidates(self):
        self.locateSymbolCandidates()
        self.locateReferenceCandidates()
        self.locatePrologueCandidates()
        self.locateLangSpecCandidates()
        self.locateStubChainCandidates()
        self.locateExceptionHandlerCandidates()
        self.identified_alignment = self._identifyAlignment()

    def _buildQueue(self):
        LOGGER.debug("Located %d function candidates", len(self.candidates))
        # increase lookup speed with static list
        self._candidate_offsets = [c.addr for c in self.candidates.values()]
        self.cached_candidates = list(self.candidates.values())
        self.candidate_queue = PriorityQueue(content=self.cached_candidates)

    def locateSymbolCandidates(self):
        for symbol_addr in self.symbol_addresses:
            self.addSymbolCandidate(symbol_addr)

    def locateReferenceCandidates(self):
        # check for potential call instructions and check if their destinations have a common function prologue
        for call_match in re.finditer(b"\xE8", self.disassembly.binary_info.binary):
            if not self._passesCodeFilter(self.disassembly.binary_info.base_addr + call_match.start()):
                continue
            if len(self.disassembly.binary_info.binary) - call_match.start() > 5:
                packed_call = self.disassembly.getRawBytes(call_match.start() + 1, 4)
                rel_call_offset = struct.unpack("i", packed_call)[0]
                # ignore zero offset calls, as they will likely not lead to functions but are rather used for positioning in shellcode etc
                if rel_call_offset == 0:
                    continue
                call_destination = (self.disassembly.binary_info.base_addr + rel_call_offset + call_match.start() + 5) & self.getBitMask()
                if self.disassembly.isAddrWithinMemoryImage(call_destination):
                    self.addReferenceCandidate(call_destination, self.disassembly.binary_info.base_addr + call_match.start())
                    self.setInitialCandidate(call_destination)
        # also check for "jmp dword ptr <offset>", as they sometimes point to local functions (i.e. non-API)
        if self.bitness == 32:
            for match in re.finditer(b"\xFF\x25", self.disassembly.binary_info.binary):
                function_addr = self.resolvePointerReference(match.start())
                if not self._passesCodeFilter(function_addr):
                    continue
                if self.disassembly.isAddrWithinMemoryImage(function_addr):
                    self.addReferenceCandidate(function_addr, self.disassembly.binary_info.base_addr + match.start())
                    self.setInitialCandidate(function_addr)
            # also check for "call dword ptr <offset>", as they sometimes point to local functions (i.e. non-API)
            for match in re.finditer(b"\xFF\x15", self.disassembly.binary_info.binary):
                function_addr = self.resolvePointerReference(match.start())
                if not self._passesCodeFilter(function_addr):
                    continue
                if self.disassembly.isAddrWithinMemoryImage(function_addr):
                    self.addReferenceCandidate(function_addr, self.disassembly.binary_info.base_addr + match.start())
                    self.setInitialCandidate(function_addr)

    def locatePrologueCandidates(self):
        # next check for the default function prologue regardless of references
        for re_prologue in DEFAULT_PROLOGUES:
            for prologue_match in re.finditer(re_prologue, self.disassembly.binary_info.binary):
                if not self._passesCodeFilter(self.disassembly.binary_info.base_addr + prologue_match.start()):
                    continue
                self.addPrologueCandidate((self.disassembly.binary_info.base_addr + prologue_match.start()) & self.getBitMask())
                self.setInitialCandidate((self.disassembly.binary_info.base_addr + prologue_match.start()) & self.getBitMask())

    def locateLangSpecCandidates(self):
        # if the sample is highly likely delphi, extract t-string-objects and use their function-addresses as high-confidence function starts
        delphi_candidates = set([])
        if self.lang_analyzer.checkDelphi():
            LOGGER.debug("Programming language recognized as Delphi, adding function start addresses from TObjects")
            t_objects = self.lang_analyzer.getDelphiObjects()
            for t_string in t_objects:
                delphi_candidates.update(set(t_objects[t_string]))
            LOGGER.debug("delphi candidates based on TObject analysis: %d", len(delphi_candidates))
            for obj in delphi_candidates:
                self.addLanguageSpecCandidate(obj, "delphi")
        if self.lang_analyzer.checkGo():
            LOGGER.debug("Programming language recognized as Go, adding function start addresses from PCLNTAB")
            self.go_objects = self.lang_analyzer.getGoObjects()
            for add in self.go_objects:
                self.addLanguageSpecCandidate(add, 'go')


    def locateStubChainCandidates(self):
        # binaries often contain long sequences of stubs, consisting only of jmp dword ptr <offset>, add such chains as candidates
        for block in re.finditer(b"(?P<block>(\xFF\x25[\S\s]{4}){2,})", self.disassembly.binary_info.binary):
            for match in re.finditer(b"\xFF\x25(?P<function>[\S\s]{4})", block.group("block")):
                stub_addr = self.disassembly.binary_info.base_addr + block.start() + match.start()
                if not self._passesCodeFilter(stub_addr):
                    continue
                self.addPrologueCandidate(stub_addr & self.getBitMask())
                self.setInitialCandidate(stub_addr & self.getBitMask())
                self.candidates[stub_addr].setIsStub(True)
        # structure for plt entries is similar but interleaved with additional code not considered functions
        for block in re.finditer(b"(?P<block>(\xFF\x25[\S\s]{4}\x68[\S\s]{4}\xE9[\S\s]{4}){2,})", self.disassembly.binary_info.binary):
            for match in re.finditer(b"\xFF\x25(?P<function>[\S\s]{4})", block.group("block")):
                stub_addr = self.disassembly.binary_info.base_addr + block.start() + match.start()
                if not self._passesCodeFilter(stub_addr):
                    continue
                self.addPrologueCandidate(stub_addr & self.getBitMask())
                self.setInitialCandidate(stub_addr & self.getBitMask())
                self.candidates[stub_addr].setIsStub(True)
                # define data bytes inbetween
                for offset in range(10):
                    self.disassembly.data_map.add(stub_addr + 6 + offset)
        # structure for plt.sec (Intel Control Flow Enforcement Technology) entries
        """
        those look e.g. like this (64bit):
        .plt.sec:000000000000CF70                                           ; =============== S U B R O U T I N E =======================================
        .plt.sec:000000000000CF70
        .plt.sec:000000000000CF70                                           ; Attributes: thunk
        .plt.sec:000000000000CF70
        .plt.sec:000000000000CF70                                           ; time_t time(time_t *timer)
        .plt.sec:000000000000CF70                                           _time           proc near               ; CODE XREF: main+BE↓p
        .plt.sec:000000000000CF70                                                                                   ; li_rand_init+37↓p ...
        .plt.sec:000000000000CF70 F3 0F 1E FA                                               endbr64
        .plt.sec:000000000000CF74 F2 FF 25 0D 2E 05 00                                      bnd jmp cs:time_ptr
        .plt.sec:000000000000CF74                                           _time           endp
        .plt.sec:000000000000CF74
        .plt.sec:000000000000CF74                                           ; ---------------------------------------------------------------------------
        .plt.sec:000000000000CF7B 0F 1F 44 00 00                                            align 20h
        """
        for block in re.finditer(b"(?P<block>(\xF3\x0F\x1E\xFA\xF2\xFF\x25[\S\s]{4}\x0F\x1F\x44\x00\x00){2,})", self.disassembly.binary_info.binary):
            for match in re.finditer(b"\xF3\x0F\x1E\xFA\xF2\xFF\x25(?P<function>[\S\s]{4})", block.group("block")):
                stub_addr = self.disassembly.binary_info.base_addr + block.start() + match.start()
                if not self._passesCodeFilter(stub_addr):
                    continue
                self.addPrologueCandidate(stub_addr & self.getBitMask())
                self.setInitialCandidate(stub_addr & self.getBitMask())
                self.candidates[stub_addr].setIsStub(True)
                # define data bytes inbetween
                for offset in range(5):
                    self.disassembly.data_map.add(stub_addr + 7 + offset)


    def locateExceptionHandlerCandidates(self):
        # 64bit only - if we have a .pdata section describing exception handlers, we extract entries of guaranteed function starts from it.
        # TODO 2020-10-29 continue here and extract function start candidates
        if self.disassembly.binary_info.bitness == 64:
            for section_info in self.disassembly.binary_info.getSections():
                section_name, section_va_start, section_va_end = section_info
                if section_name == ".pdata":
                    rva_start = section_va_start - self.disassembly.binary_info.base_addr
                    rva_end = section_va_end - self.disassembly.binary_info.base_addr
                    for offset in range(rva_start, rva_end + 1, 12):
                        packed_dword = self.disassembly.binary_info.binary[offset:offset + 4]
                        rva_function_candidate = None
                        if len(packed_dword) == 4:
                            rva_function_candidate = struct.unpack("I", packed_dword)[0]
                            self.addExceptionCandidate(self.disassembly.binary_info.base_addr + rva_function_candidate)
                        if not rva_function_candidate:
                            break



import os
import logging


class SmdaConfig(object):

    # note to self: always change this in setup.py as well!
    VERSION = "1.8.1"
    CONFIG_FILE_PATH = str(os.path.abspath(__file__))
    PROJECT_ROOT = str(os.path.abspath(os.sep.join([CONFIG_FILE_PATH, "..", ".."])))

    ### An (optional) WinAPI database as generated by ApiScout (https://github.com/danielplohmann/apiscout)
    API_COLLECTION_FILES = {}
    ### global logging-config setup
    # Only do basicConfig if no handlers have been configured
    LOG_PATH = "./"
    LOG_LEVEL = logging.INFO
    LOG_FORMAT = "%(asctime)-15s: %(name)-32s - %(message)s"

    ### SMDA disassembler config
    # maximum time in seconds for disassembly to complete
    TIMEOUT = 300
    # maximum number of bytes to allocate while loading
    MAX_IMAGE_SIZE = 100 * 1024 * 1024
    # store raw binary buffer in SmdaReport to enable carving data from refs
    STORE_BUFFER = False
    # improve disassembly by resolving references through data flows
    USE_ALIGNMENT = True
    USE_SYMBOLS_AS_CANDIDATES = True
    RESOLVE_REGISTER_CALLS = True
    HIGH_ACCURACY = True
    RESOLVE_TAILCALLS = False
    # optional metadata generation options
    CALCULATE_SCC = True
    CALCULATE_NESTING = True
    # confidence score to use for filtering functions before including them in the output
    CONFIDENCE_THRESHOLD = 0.0





class CodeXref(object):

    def __init__(self, smda_ins_from, smda_ins_to):
        self.smda_ins_from = smda_ins_from
        self.smda_ins_to = smda_ins_to

    @property
    def from_function(self):
        return self.smda_ins_from.smda_function

    @property
    def to_function(self):
        return self.smda_ins_to.smda_function

    @property
    def from_instruction(self):
        return self.smda_ins_from

    @property
    def to_instruction(self):
        return self.smda_ins_to

    def __str__(self):
        return "0x%x (0x%x) -> 0x%x (0x%x)" % (self.smda_ins_from.offset, self.smda_ins_from.smda_function.offset, self.smda_ins_to.offset, self.smda_ins_to.smda_function.offset)

    def __repr__(self):
        return "<CodeXref(SmdaInstruction({}), SmdaInstruction({}))>".format(self.smda_ins_from.offset, self.smda_ins_to.offset)



class BackendInterface(object):

    def __init__(self):
        pass

    def getArchitecture(self):
        raise NotImplementedError

    def getBitness(self):
        raise NotImplementedError

    def getFunctions(self):
        raise NotImplementedError

    def getBlocks(self, function_offset):
        raise NotImplementedError

    def getCodeInRefs(self, offset):
        raise NotImplementedError

    def getCodeOutRefs(self, offset):
        raise NotImplementedError

    def getInstructionBytes(self, offset):
        raise NotImplementedError

    def getFunctionSymbols(self):
        raise NotImplementedError

    def getBaseAddr(self):
        raise NotImplementedError

    def getBinary(self):
        raise NotImplementedError

    def getApiOffsets(self):
        raise NotImplementedError



import re

from .BackendInterface import BackendInterface

try:
    import idaapi
    import idautils
except:
    pass

try:
    # we only need these when we are in IDA - IDA 7.4 and above
    import ida_idaapi
    import ida_funcs
    import ida_gdl
    import ida_bytes
    import ida_nalt
    import ida_segment
    import ida_name
except:
    pass

try:
    # we only need these when we are in IDA - IDA 7.3 and below
    import idc
except:
    pass


class IdaInterface(object):
    # derived from https://python-3-patterns-idioms-test.readthedocs.io/en/latest/Singleton.html
    instance = None
    def __init__(self):
        if not IdaInterface.instance:
            if idaapi.IDA_SDK_VERSION >= 740:
                IdaInterface.instance = Ida74Interface()
            else:
                IdaInterface.instance = Ida73Interface()

    def __getattr__(self, name):
        return getattr(self.instance, name)

    def getIdbDir(self):
        return idautils.GetIdbDir()


class Ida74Interface(BackendInterface):

    def __init__(self):
        self.version = "IDA Pro 7.4"
        self._processor_map = {
            "metapc": "intel"
        }
        self._api_map = {}
        self._import_module_name = ""

    def getArchitecture(self):
        # https://reverseengineering.stackexchange.com/a/11398
        info = ida_idaapi.get_inf_structure()
        procname = info.procName
        if procname in self._processor_map:
            return self._processor_map[procname]
        else:
            raise ValueError("Unsupported Architecture")

    def getBitness(self):
        # https://reverseengineering.stackexchange.com/a/11398
        bits = None
        info = ida_idaapi.get_inf_structure()
        if info.is_64bit():
            bits = 64
        elif info.is_32bit():
            bits = 32
        else:
            bits = 16
        return bits

    def getFunctions(self):
        return sorted([offset for offset in idautils.Functions()])

    def getBlocks(self, function_offset):
        blocks = []
        function_chart = ida_gdl.FlowChart(ida_funcs.get_func(function_offset))
        for block in function_chart:
            extracted_block = []
            for instruction in idautils.Heads(block.start_ea, block.end_ea):
                if ida_bytes.is_code(ida_bytes.get_flags(instruction)):
                    extracted_block.append(instruction)
            if extracted_block:
                blocks.append(extracted_block)
        return sorted(blocks)

    def getInstructionBytes(self, offset):
        ins = idautils.DecodeInstruction(offset)
        ins_bytes = ida_bytes.get_bytes(offset, ins.size)
        return ins_bytes

    def getCodeInRefs(self, offset):
        return [(ref_from, offset) for ref_from in idautils.CodeRefsTo(offset, True)]

    def getCodeOutRefs(self, offset):
        return [(offset, ref_to) for ref_to in idautils.CodeRefsFrom(offset, True)]

    def getFunctionSymbols(self):
        function_symbols = {}
        function_offsets = self.getFunctions()
        for function_offset in function_offsets:
            function_name = ida_funcs.get_func_name(function_offset)
            # apply demangling if required
            if "@" in function_name:
                demangled = ida_name.demangle_name(function_name, 0)
                if demangled:
                    function_name = demangled
            if not re.match("sub_[0-9a-fA-F]+", function_name):
                function_symbols[function_offset] = function_name
        return function_symbols

    def getBaseAddr(self):
        base_addr = 0
        segment_starts = [ea for ea in idautils.Segments()]
        if segment_starts:
            first_segment_start = segment_starts[0]
            # re-align by 0x10000 to reflect typically allocation behaviour for IDA-mapped binaries
            first_segment_start = (first_segment_start / 0x10000) * 0x10000
            base_addr = int(first_segment_start)
        return base_addr

    def getBinary(self):
        result = b""
        segment = ida_segment.get_first_seg()
        while segment:
            result += ida_bytes.get_bytes(segment.start_ea, segment.end_ea - segment.start_ea)
            segment = ida_segment.get_next_seg(segment.end_ea)
        return result

    def getApiMap(self):
        self._api_map = {}
        num_imports = ida_nalt.get_import_module_qty()
        for i in range(0, num_imports):
            self._import_module_name = ida_nalt.get_import_module_name(i)
            ida_nalt.enum_import_names(i, self._cbEnumImports)
        return self._api_map

    def isExternalFunction(self, function_offset):
        function_segment = ida_segment.getseg(function_offset)
        function_segment_name = ida_segment.get_segm_name(function_segment)
        is_extern = function_segment_name in ["extern", "UNDEF"]
        return is_extern

    def _cbEnumImports(self, addr, name, ordinal):
        # potentially use: idc.Name(addr)
        if self._import_module_name:
            self._api_map[addr] = self._import_module_name + "!" + name
        else:
            self._api_map[addr] = name
        return True



class Ida73Interface(BackendInterface):

    def __init__(self):
        self.version = "IDA Pro 7.3 and below"
        self._processor_map = {
            "metapc": "intel"
        }
        self._api_map = {}
        self._import_module_name = ""

    def getArchitecture(self):
        # https://reverseengineering.stackexchange.com/a/11398
        info = idaapi.get_inf_structure()
        procname = info.procName
        if procname in self._processor_map:
            return self._processor_map[procname]
        else:
            raise ValueError("Unsupported Architecture")

    def getBitness(self):
        # https://reverseengineering.stackexchange.com/a/11398
        bits = None
        info = idaapi.get_inf_structure()
        if info.is_64bit():
            bits = 64
        elif info.is_32bit():
            bits = 32
        else:
            bits = 16
        return bits

    def getFunctions(self):
        return sorted([offset for offset in idautils.Functions()])

    def getBlocks(self, function_offset):
        blocks = []
        function_chart = idaapi.FlowChart(idaapi.get_func(function_offset))
        for block in function_chart:
            extracted_block = []
            for instruction in idautils.Heads(block.startEA, block.endEA):
                if idc.isCode(idc.GetFlags(instruction)):
                    extracted_block.append(instruction)
            if extracted_block:
                blocks.append(extracted_block)
        return sorted(blocks)

    def getInstructionBytes(self, offset):
        ins = idautils.DecodeInstruction(offset)
        ins_bytes = idc.get_bytes(offset, ins.size)
        return ins_bytes

    def getCodeInRefs(self, offset):
        return [(ref_from, offset) for ref_from in idautils.CodeRefsTo(offset, True)]

    def getCodeOutRefs(self, offset):
        return [(offset, ref_to) for ref_to in idautils.CodeRefsFrom(offset, True)]

    def getFunctionSymbols(self):
        function_symbols = {}
        function_offsets = self.getFunctions()
        for function_offset in function_offsets:
            function_name = idc.GetFunctionName(function_offset)
            # apply demangling if required
            if "@" in function_name:
                function_name = idc.demangle_name(function_name, 0)
            if not re.match("sub_[0-9a-fA-F]+", function_name):
                function_symbols[function_offset] = function_name
        return function_symbols

    def getBaseAddr(self):
        segment_starts = [ea for ea in idautils.Segments()]
        first_segment_start = segment_starts[0]
        # re-align by 0x10000 to reflect typically allocation behaviour for IDA-mapped binaries
        first_segment_start = (first_segment_start / 0x10000) * 0x10000
        return int(first_segment_start)

    def getBinary(self):
        result = b""
        segment_starts = [ea for ea in idautils.Segments()]
        offsets = []
        start_len = 0
        for start in segment_starts:
            end = idc.SegEnd(start)
            result += idc.get_bytes(start, end - start)
            offsets.append((start, start_len, len(result)))
            start_len = len(result)
        return result

    def getApiMap(self):
        self._api_map = {}
        num_imports = idaapi.get_import_module_qty()
        for i in range(0, num_imports):
            self._import_module_name = idaapi.get_import_module_name(i)
            idaapi.enum_import_names(i, self._cbEnumImports)
        return self._api_map

    def isExternalFunction(self, function_offset):
        # TODO look up older function names to support this for IDA 7.3- as well
        return False

    def _cbEnumImports(self, addr, name, ordinal):
        # potentially use: idc.Name(addr)
        if self._import_module_name:
            self._api_map[addr] = self._import_module_name + "!" + name
        else:
            self._api_map[addr] = name
        return True



import os
from smda.utility.PeFileLoader import PeFileLoader
from smda.utility.ElfFileLoader import ElfFileLoader


class FileLoader(object):

    _file_path = None
    _map_file = False
    _data = b""
    _raw_data = b""
    _base_addr = 0
    _bitness = 0
    _code_areas = []
    file_loaders = [PeFileLoader, ElfFileLoader]

    def __init__(self, file_path, load_file=True, map_file=False):
        self._file_path = file_path
        self._map_file = map_file
        if load_file:
            self._loadFile()

    def _loadRawFileContent(self):
        binary = ""
        if os.path.isfile(self._file_path):
            with open(self._file_path, "rb") as inf:
                binary = inf.read()
        return binary

    def _loadFile(self, buffer=None):
        self._raw_data = buffer if buffer is not None else self._loadRawFileContent()
        if self._map_file:
            for loader in self.file_loaders:
                if loader.isCompatible(self._raw_data):
                    self._data = loader.mapBinary(self._raw_data)
                    self._base_addr = loader.getBaseAddress(self._raw_data)
                    self._bitness = loader.getBitness(self._raw_data)
                    self._code_areas = loader.getCodeAreas(self._raw_data)
                    break
        else:
            self._data = self._raw_data

    def getData(self):
        return self._data

    def getRawData(self):
        return self._raw_data

    def getBaseAddress(self):
        return self._base_addr

    def getBitness(self):
        return self._bitness

    def getCodeAreas(self):
        return self._code_areas



# Implementation by Armin Rigo
# source: https://bitbucket.org/arigo/arigo/src/default/hack/pypy-hack/heapstats/dominator.py

# Implementation following:
#
#   Lengauer, Thomas; and Tarjan, Robert Endre (July 1979).
#   "A fast algorithm for finding dominators in a flowgraph".
#   ACM Transactions on Programming Languages and Systems (TOPLAS) 1 (1):
#   121-141.
#
#   http://portal.acm.org/ft_gateway.cfm?id=357071

import logging

LOGGER = logging.getLogger(__name__)

class DominatorTree(object):

    def __init__(self, G, r):
        assert r in G
        self.succ = G
        self.r = r

    def init_variables(self):
        self.parent = {}
        self.pred = {}
        self.semi = {}
        self.vertex = []
        self.bucket = {}
        self.dom = {}
        self.ancestor = {}
        self.label = {}
        for v in self.succ:
            self.pred[v] = set()
            self.bucket[v] = set()

    def depth_first_search(self, v):
        stack = [v]
        while stack:
            v = stack.pop()
            n = len(self.vertex)
            self.semi[v] = n
            self.vertex.append(v)
            for w in self.succ[v]:
                self.pred[w].add(v)
                if w not in self.semi:
                    self.parent[w] = v
                    self.semi[w] = None     # temporarily
                    stack.append(w)

    def LINK(self, v, w):
        self.ancestor[w] = v

    def EVAL(self, v):
        if v not in self.ancestor:
            return v
        else:
            self.COMPRESS(v)
            return self.label.get(v, v)

    def COMPRESS(self, v):
        if self.ancestor[v] in self.ancestor:
            self.COMPRESS(self.ancestor[v])
            w = self.ancestor[v]
            if self.semi[self.label.get(w,w)] < self.semi[self.label.get(v,v)]:
                self.label[v] = self.label.get(w,w)
            self.ancestor[v] = self.ancestor[w]

    def steps_2_3(self):
        for w in self.vertex[:0:-1]:
            # step 2
            for v in self.pred[w]:
                u = self.EVAL(v)
                if self.semi[u] < self.semi[w]:
                    self.semi[w] = self.semi[u]
            self.bucket[self.vertex[self.semi[w]]].add(w)
            self.LINK(self.parent[w], w)
            # step 3
            for v in list(self.bucket[self.parent[w]]):
                self.bucket[self.parent[w]].remove(v)
                u = self.EVAL(v)
                if self.semi[u] < self.semi[v]:
                    self.dom[v] = u
                else:
                    self.dom[v] = self.parent[w]

    def step_4(self):
        for w in self.vertex[1:]:
            if self.dom[w] != self.vertex[self.semi[w]]:
                self.dom[w] = self.dom[self.dom[w]]

    def compute(self):
        self.init_variables()
        self.depth_first_search(self.r)
        self.steps_2_3()
        self.step_4()


def fix_graph(graph):
    expanded_graph = {}
    for key, values in graph.items():
        expanded_graph[key] = values
        for value in values:
            if value not in expanded_graph:
                expanded_graph[value] = []
    return expanded_graph


# Calculation of Nesting Depth by walking down dominators and summarizing weights
# Implementation by Steffen Enders and Daniel Plohmann

def build_dominator_tree(G, r):
    expanded_graph = fix_graph(G)
    if not r in expanded_graph:
        # print("r not in G:", r, G)
        LOGGER.debug("r not in G: %s %s", r, G)
        return None
    domtree = DominatorTree(expanded_graph, r)
    domtree.compute()
    inverted = {}
    for key, value in domtree.dom.items():
        if value not in inverted:
            inverted[value] = []
        inverted[value].append(key)
    return inverted

def get_nesting_depth(graph, domtree, root):
    expanded_graph = fix_graph(graph)
    significant_nodes = set.union(*([set(v) for v in expanded_graph.values() if len(v) > 1] + [set()]))
    # print("significant_nodes", significant_nodes)
    def maximum_costs(cn):
        # print("  maximum_costs cn", cn)
        if cn not in domtree or not domtree[cn]:
            # print("    %d not in domtree or not domtree[%d]" % (cn, cn), 1 if cn in significant_nodes else 0)
            return (1 if cn in significant_nodes else 0)
        val = max(maximum_costs(n) for n in domtree[cn]) + (1 if cn in significant_nodes else 0)
        # print("   ", val, 1 if cn in significant_nodes else 0)
        return val
    try:
        return maximum_costs(root)
    except:
        return 0



if __name__ == "__main__":
    test_data = [
        {
            "smda": {10208: [10229], 10229: [10240, 10253], 10240: [10244, 10246], 10244: [10246], 10246: [10240, 10253], 10253: [10229, 10261]},
            "smda_function": 10208,
            "fixed": {10208: [10229], 10229: [10240, 10253], 10240: [10244, 10246], 10253: [10229, 10261], 10244: [10246], 10246: [10240, 10253], 10261: []},
            "dt": {10240: [10244, 10246], 10229: [10240, 10253], 10253: [10261], 10208: [10229]},
            "nd": 3
        }, {
            "smda": {1: [2], 2: [3, 4, 6], 3: [5], 4: [5], 5: [2]},
            "smda_function": 1,
            "fixed": {1: [2], 2: [3, 4, 6], 3: [5], 4: [5], 6: [], 5: [2]},
            "dt": {2: [3, 4, 5, 6], 1: [2]},
            "nd": 1
        }, {
            "smda": {1: [2], 2: [3, 6], 3: [41, 42], 41: [5], 42: [5], 5: [2]},
            "smda_function": 1,
            "fixed": {1: [2], 2: [3, 6], 3: [41, 42], 6: [], 41: [5], 42: [5], 5: [2]},
            "dt": {3: [41, 42, 5], 2: [3, 6], 1: [2]},
            "nd": 2
        },

    ]
    for data in test_data:
        print("*" * 80)
        print("Running Test Case: ", data["smda_function"])
        print("*" * 80)
        print("smda", data["smda"])
        fixed_smda = {}
        for key, values in data["smda"].items():
            fixed_smda[key] = values
            for value in values:
                if value not in fixed_smda:
                    fixed_smda[value] = []
        print("fixed_smda", fixed_smda)
        assert fixed_smda == data["fixed"]
        dt = build_dominator_tree(data["smda"], data["smda_function"])
        print("dominator tree", dt)
        assert dt == data["dt"]
        nd = get_nesting_depth(fixed_smda, dt, data["smda_function"])
        print("nd", nd)
        assert nd == data["nd"]



"""
    Tarjan's Algorithm (named for its discoverer, Robert Tarjan) is a graph theory algorithm
    for finding the strongly connected components of a graph.
    This can be used to find loops.
    Based on: http://en.wikipedia.org/wiki/Tarjan%27s_strongly_connected_components_algorithm
    - Refactored into a class to allow pooled computation by Daniel Plohmann
    - Implementation by Bas Westerbaan:
      https://github.com/bwesterb/py-tarjan
"""

class Tarjan(object):
    """ g is the graph represented as a dictionary { <vertex> : <successors of vertex> } """

    def __init__(self, graph):
        self._graph = graph
        self._stack = []
        self._stack_set = set([])
        self._index = {}
        self._lowlink = {}
        self._nonrecursive_stack = []
        self._result = []

    def _tarjan_head(self, v):
        self._index[v] = len(self._index)
        self._lowlink[v] = self._index[v]
        self._stack.append(v)
        self._stack_set.add(v)
        it = iter(self._graph.get(v, ()))
        self._nonrecursive_stack.append((it, False, v, None))

    def _tarjan_body(self, it, v):
        for w in it:
            if w not in self._index:
                self._nonrecursive_stack.append((it, True, v, w))
                self._tarjan_head(w)
                return
            if w in self._stack_set:
                self._lowlink[v] = min(self._lowlink[v], self._index[w])
        if self._lowlink[v] == self._index[v]:
            scc = []
            w = None
            while v != w:
                w = self._stack.pop()
                scc.append(w)
                self._stack_set.remove(w)
            self._result.append(scc)

    def calculateScc(self):
        main_iter = iter(self._graph)
        while True:
            try:
                v = next(main_iter)
            except StopIteration:
                return self._result
            if v not in self._index:
                self._tarjan_head(v)
            while self._nonrecursive_stack:
                it, inside, v, w = self._nonrecursive_stack.pop()
                if inside:
                    self._lowlink[v] = min(self._lowlink[w], self._lowlink[v])
                self._tarjan_body(it, v)

    def closure(self):
        """ Given a graph @g, returns the transitive closure of @g """
        ret = {}
        for scc in self.calculateScc():
            ws = set()
            ews = set()
            for v in scc:
                ws.update(self._graph[v])
            for w in ws:
                assert w in ret or w in scc
                ews.add(w)
                ews.update(ret.get(w, ()))
            if len(scc) > 1:
                ews.update(scc)
            ews = tuple(ews)
            for v in scc:
                ret[v] = ews
        return ret

    def getResult(self):
        return self._result



#!/usr/bin/python

from abc import abstractmethod

import logging
LOGGER = logging.getLogger(__name__)


class AbstractLabelProvider:

    def __init__(self, config):
        raise NotImplementedError

    @abstractmethod
    def update(self, binary_info):
        """If the LabelProvider needs to parse from the given target, update() can be used to populate the provider """
        raise NotImplementedError

    @abstractmethod
    def getApi(self, absolute_addr):
        """If the LabelProvider has any information about a used API for the given address, return (dll, api), else return None"""
        raise NotImplementedError

    @abstractmethod
    def getSymbol(self, address):
        """If the LabelProvider has any information about a used Symbol for the given address, return the symbol, else return None"""
        raise NotImplementedError

    @abstractmethod
    def isApiProvider(self):
        """Returns whether the get_api(..) function of the AbstractLabelProvider is functional"""
        return False

    @abstractmethod
    def isSymbolProvider(self):
        """Returns whether the get_symbol(..) function of the AbstractLabelProvider is functional"""
        return False

    @abstractmethod
    def getFunctionSymbols(self):
        """Return all function symbol data """
        return {}



from smda.intel.IntelInstructionEscaper import IntelInstructionEscaper
from capstone.x86 import X86_OP_IMM, X86_OP_MEM

class SmdaInstruction:

    smda_function = None
    offset = None
    bytes = None
    mnemonic = None
    operands = None
    detailed = None

    def __init__(self, ins_list=None, smda_function=None):
        self.smda_function = smda_function
        if ins_list is not None:
            self.offset = ins_list[0]
            self.bytes = ins_list[1]
            self.mnemonic = ins_list[2]
            self.operands = ins_list[3]

    def getDataRefs(self):
        if self.getMnemonicGroup(IntelInstructionEscaper) != "C":
            detailed = self.getDetailed()
            if len(detailed.operands) > 0:
                for i in detailed.operands:
                    value = None
                    if i.type == X86_OP_IMM:
                        value = i.imm
                    if i.type == X86_OP_MEM:
                        value = i.mem.disp
                        if detailed.reg_name(i.mem.base) == "rip":
                            # add RIP value
                            value += detailed.address + detailed.size
                    if value is not None and self.smda_function.smda_report.isAddrWithinMemoryImage(value):
                        yield value

    def getDetailed(self):
        if self.detailed is None:
            capstone = self.smda_function.smda_report.getCapstone()
            with_details = [i for i in capstone.disasm(bytes.fromhex(self.bytes), self.offset)]
            assert len(with_details) == 1
            self.detailed = with_details[0]
        return self.detailed

    def getMnemonicGroup(self, escaper):
        if escaper:
            return escaper.escapeMnemonic(self.mnemonic)
        return self.bytes

    def getEscapedOperands(self, escaper):
        if escaper:
            return escaper.escapeOperands(self)
        return self.bytes

    def getMaskedOperands(self, escaper):
        if escaper:
            return escaper.escapeOperands(self, offsets_only=True)
        return self.bytes

    def getEscapedBinary(self, escaper, lower_addr=None, upper_addr=None):
        if escaper:
            return escaper.escapeBinary(self, lower_addr=lower_addr, upper_addr=upper_addr)
        return self.bytes

    @classmethod
    def fromDict(cls, instruction_dict, smda_function=None):
        smda_instruction = cls(None)
        smda_instruction.smda_function = smda_function
        smda_instruction.offset = instruction_dict[0]
        smda_instruction.bytes = instruction_dict[1]
        smda_instruction.mnemonic = instruction_dict[2]
        smda_instruction.operands = instruction_dict[3]
        return smda_instruction

    def toDict(self):
        return [self.offset, self.bytes, self.mnemonic, self.operands]

    def __int__(self):
        return self.offset

    def __str__(self):
        return "0x{:08x}: ({:>14s}) - {} {}".format(self.offset, self.bytes, self.mnemonic, self.operands)



from collections import defaultdict
from operator import itemgetter
import bisect
import json

class TailcallAnalyzer(object):

    def __init__(self):
        self.__jumps = defaultdict(set)
        self.__tmp_jumps = defaultdict(list)
        self.__functions = list()

    def initFunction(self):
        self.__tmp_jumps = defaultdict(list)

    def addJump(self, source, destination):
        self.__tmp_jumps[source].append(destination)

    def finalizeFunction(self, function_state):
        for source, destinations in self.__tmp_jumps.items():
            self.__jumps[source].update(destinations)
        self.__tmp_jumps.clear()
        self.__functions.append(function_state)

    def getTailcalls(self):
        result = list()
        # jumps sorted by (destination, source)
        jumps = list(sorted(((s, d) for s in self.__jumps for d in self.__jumps[s]), key=itemgetter(1, 0)))
        jumps_dest = [d for s, d in jumps]
        # for each function generate the intervals that contain the instructions
        for function in self.__functions:
            # check if there are any jumps from outside the function to inside the function
            function_intervals = self.__getFunctionIntervals(function)
            if not function_intervals:
                # empty function?
                continue
            min_addr = min(interval[0] for interval in function_intervals)
            max_addr = max(interval[1] for interval in function_intervals)
            for source, destination in jumps[bisect.bisect_left(jumps_dest, min_addr):bisect.bisect_right(jumps_dest, max_addr)]:
                if (
                        # the jumps destination is different from the functions start address
                        destination != function.start_addr and
                        # the jumps destination is in one of the functions intervals
                        any((first <= destination <= last) for first, last in function_intervals) and
                        # the jump originates from outside the function (outside all intervals)
                        all((source < first or source > last) for first, last in function_intervals)):

                    result.append({
                        "source_addr": source,
                        "destination_addr": destination,
                        "destination_function": function.start_addr
                    })

        return result

    def __getFunctionIntervals(self, function_state):
        intervals = list()
        instructions = sorted(function_state.instructions, key=itemgetter(0))
        first_instruction = instructions[0] if instructions else None
        last_instruction = first_instruction
        for instruction in instructions:
            if instruction[0] > last_instruction[0] + last_instruction[1]:
                intervals.append((first_instruction[0], last_instruction[0]))
                first_instruction = instruction
            last_instruction = instruction
        if last_instruction:
            intervals.append((first_instruction[0], last_instruction[0]))
        return intervals

    def __getFunctionByStartAddr(self, start_addr):
        for function in self.__functions:
            if function.start_addr == start_addr:
                return function

    def __printIntervals(self, intervals):
        # return
        if len(intervals) < 25:
            for one, two in intervals:
                print("  0x{:x} -> 0x{:x}".format(one, two))
        else: print("Function has too many intervals to display")

    def resolveTailcalls(self, disassembler, verbose=False):
        newly_created_functions = set([])
        for tailcall in self.getTailcalls():
            if verbose:
                print("Processing tailcall:\n{}".format(json.dumps(tailcall, indent=2, sort_keys=True)))
            # remove the information from the function-analysis state of the disassembly
            function = self.__getFunctionByStartAddr(tailcall["destination_function"])
            if not function or function.is_tailcall_function:
                disassembler.analyzeFunction(tailcall["destination_function"])
                continue

            self.__functions.remove(function)
            if function:
                if verbose:
                    print("Old function:")
                    self.__printIntervals(self.__getFunctionIntervals(function))
                function.revertAnalysis()

            # analyze the tailcall destination as function
            disassembler.analyzeFunction(tailcall["destination_addr"])
            newly_created_functions.add(tailcall["destination_addr"])
            function = self.__getFunctionByStartAddr(tailcall["destination_addr"])
            if function and not tailcall["destination_function"] in function.instruction_start_bytes:
                # analyze the (previously) broken function a second time
                try:
                    disassembler.analyzeFunction(tailcall["destination_function"])
                    function = self.__getFunctionByStartAddr(tailcall["destination_function"])
                    function.is_tailcall_function = True
                except:
                    pass
                    # print ("0x{:x} -> 0x{:x}".format(tailcall["destination_function"], tailcall["destination_addr"]))
            elif verbose:
                print("**** 0x{:x} IS NOW PART OF 0x{:x}".format(tailcall["destination_function"], tailcall["destination_addr"]))

            if verbose:
                function = self.__getFunctionByStartAddr(tailcall["destination_function"])
                new_function = self.__getFunctionByStartAddr(tailcall["destination_addr"])
                print("New function:")
                if new_function:
                    self.__printIntervals(self.__getFunctionIntervals(new_function))
                print("Re-disassembled old function:")
                if function:
                    self.__printIntervals(self.__getFunctionIntervals(function))
        return sorted(list(newly_created_functions))



class OrdinalHelper(object):
    # TODO POC implementation, extend list. ole32.dll and mfc42.dll are candidates here
    ORDINALS = {
        "ws2_32.dll": {
            1: "accept",
            2: "bind",
            3: "closesocket",
            4: "connect",
            97: "freeaddrinfo",
            98: "getaddrinfo",
            99: "getnameinfo",
            51: "gethostbyaddr",
            52: "gethostbyname",
            53: "getprotobyname",
            54: "getprotobynumber",
            55: "getservbyname",
            56: "getservbyport",
            57: "gethostname",
            5: "getpeername",
            6: "getsockname",
            7: "getsockopt",
            8: "htonl",
            9: "htons",
            10: "ioctlsocket",
            11: "inet_addr",
            12: "inet_ntoa",
            13: "listen",
            14: "ntohl",
            15: "ntohs",
            16: "recv",
            17: "recvfrom",
            18: "select",
            19: "send",
            20: "sendto",
            21: "setsockopt",
            22: "shutdown",
            23: "socket"
        }
    }

    @staticmethod
    def resolveOrdinal(dll_name, ordinal):
        dll_name = dll_name.lower()
        if dll_name in OrdinalHelper.ORDINALS and ordinal in OrdinalHelper.ORDINALS[dll_name]:
            return OrdinalHelper.ORDINALS[dll_name][ordinal]
        return ""



#!/usr/bin/python

import logging
from .AbstractLabelProvider import AbstractLabelProvider

LOGGER = logging.getLogger(__name__)

try:
    import lief
    lief.logging.disable()
except:
    lief = None
    LOGGER.warning("3rd party library LIEF not installed - won't be able to extract symbols for ELF files where available.")



class ElfSymbolProvider(AbstractLabelProvider):
    """ Minimal resolver for ELF symbols """

    def __init__(self, config):
        self._config = config
        #addr:func_name
        self._func_symbols = {}

    def isSymbolProvider(self):
        return True

    def _parseOep(self, lief_result):
        if lief_result:
            self._func_symbols[lief_result.header.entrypoint] = "original_entry_point"

    def update(self, binary_info):
        #works both for PE and ELF
        self._func_symbols = {}
        if not binary_info.file_path:
            return
        data = ""
        with open(binary_info.file_path, "rb") as fin:
            data = fin.read(16)
        if data[:4] != b"\x7FELF" or lief is None:
            return
        lief_binary = lief.parse(binary_info.file_path)
        self._parseOep(lief_binary)
        # TODO split resolution into API/dynamic part and local symbols
        self._parseExports(lief_binary)
        self._parseSymbols(lief_binary.static_symbols)
        self._parseSymbols(lief_binary.dynamic_symbols)
        for reloc in lief_binary.relocations:
            if reloc.has_symbol:
                self._func_symbols[reloc.address] = reloc.symbol.name

    def _parseExports(self, binary):
        for function in binary.exported_functions:
            self._func_symbols[function.address] = function.name

    def _parseSymbols(self, symbols):
        for symbol in symbols:
            if symbol.is_function:
                if symbol.value != 0:
                    func_name = ""
                    try:
                        func_name = symbol.demangled_name
                    except:
                        func_name = symbol.name
                    self._func_symbols[symbol.value] = func_name

    def getSymbol(self, address):
        return self._func_symbols.get(address, "")

    def getFunctionSymbols(self):
        return self._func_symbols
